/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.chainofresponsibility;

import dk.rode.thesis.flyweight.Character;
import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.strategy.ObjectPolicy;

/**
  * An abstract <i>character handler</i> implements the basic traits
  * of handlers handling {@link Character} objects. <p>
  * 
  * Character handlers ignore null requests. 
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  */
@Participant("Handler")
public abstract class CharacterHandler implements Handler<Character> {

    /**
      * No-arg constructor. <p>
      * 
      * This handler is not registered to any 
      * handler {@link HandlerChain chain} until specifically 
      * registered using the {@link #register(HandlerChain)} 
      * method.
      */
    protected CharacterHandler() {
        super();
    }
    
    /**
      * Constructor, which registers this character handler to the
      * end of the handler chain supplied as {@code chain}.
      * 
      * @param  chain The handler chain; cannot be null.
      * @throws NullPointerException If {@code chain} is null.
      */
    protected CharacterHandler(HandlerChain<Character> chain) {
        this.register(chain);
    }
    
    /**
      * Registers this character handler to the handler chain supplied 
      * as {@code chain}.
      * 
      * @param  chain The handler chain; cannot be null.
      * @throws NullPointerException If {@code chain} is null.
      * @throws IllegalArgumentException If this handler has already been
      *         registered to {@code chain}.
      */
    public void register(HandlerChain<Character> chain) {
        chain.register(this);
    }

    /**
      * Implements the semantics required for chaining character 
      * handlers so sub-classes do not have to. The actual handling is 
      * performed by the {@link #handle(Character)} method, 
      * which sub-classes must implement. <p>
      * 
      * If this handler can handle {@code character}, this
      * handler is returned. If not, {@code link} is used
      * to forward {@code character} to the next handler
      * in the current chain. If {@code link} is null,  null
      * is returned.  
      */
    public Handler<Character> handle(Character character, HandlerLink<Character> link) {
        return (this.handle(character)) ? this : link == null ? null : link.forward(character, this);
    }
    
    /**
      * Returns true if this handler can and have handled 
      * {@code character}, false if not.
      *  
      * @param  character The character; can be null. 
      * @return True if {@code character} was handled, false if not.
      */
    protected abstract boolean handle(Character character);
    
    /**
      * Returns the string representation of this handler.
      * 
      * @return The string representation; never null.
      * @see    ObjectPolicy#ID
      */
    public String toString() {
        return ObjectPolicy.ID.asString(this);
    }
    
}
