/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 * 
 */
package dk.rode.thesis.chainofresponsibility;

import java.util.ArrayList;
import java.util.List;

import dk.rode.thesis.facade.MathFacade;
import dk.rode.thesis.flyweight.Character;
import dk.rode.thesis.flyweight.CharacterFactory;
import dk.rode.thesis.flyweight.Sentence;
import dk.rode.thesis.flyweight.Word;
import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.log.Log;
import dk.rode.thesis.meta.log.LogFactory;
import dk.rode.thesis.meta.test.Testable;
import dk.rode.thesis.meta.util.Arguments;

/**
  * <b>Chain of Responsibility</b> tests.
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  */
@Participant("Client")
public class Main implements Testable {

    /**
      * Executes the <b>Chain of Responsibility</b> tests.
      * 
      * @param args The arguments to supply to {@link #test(Log, Arguments)},
      *             if any; can be null.
      */
    public static void main(String[] args) {
        new Main().test(LogFactory.getLog(), new Arguments(args));
    }

    /**
      * No-arg constructor.
      */
    public Main() {
        super();
    }
    
    /**
      * 
      */
    public boolean test(final Log out, Arguments arguments) {
        boolean ok = true;
        
        out.heading("Chain of Responsibility tests...");

        if (arguments.size() > 0) {
            out.println("Supplied arguments: ", arguments);
        }
        
        boolean log = arguments.getArgumentValue("log", false, Boolean.class);
        LogFactory.getLog(StandardHandlerChain.class).setActive(log);
        LogFactory.getLog(WeakHandlerChain.class).setActive(log);
        LogFactory.getLog(Handler.class).setActive(true);
        LogFactory.getLog(HandlerLink.class).setActive(true);
        
        //
        // Create chain and handlers
        //
        
        // Handler chain to handle Character requests...
        final HandlerChain<Character> chain1 = new StandardHandlerChain<Character>();

        // Create and register automatically: constructor calls
        // register method on chain...
        CharacterHandler ch = new WhitespaceHandler(chain1);
        new SymbolHandler(chain1);
        // Only handles lower case letters...
        new LetterCaseHandler(chain1, false);

        final Handler<Character> sink = new Handler<Character>() {
            public Handler<Character> handle(Character request, HandlerLink<Character> link) {
                // We did not handle it, we just log it. Hence, several handlers 
                // can actually handle the same request by returning null...!
                LogFactory.getLog(this).warn("Down the drain: ", request);
                return null;
            }
            public String toString() {
                return "Sink";
            }
        };
        // Explicit link usage as alternative to chains...
        HandlerLink<Character> linkToSink = new HandlerLink<Character>() {
            public Handler<Character> forward(Character request, Handler<Character> handler) {
                LogFactory.getLog(this).println("Using explicit link to: ", sink);
                return sink.handle(request, null);
            }
        };

        // Chain to handle all...
        HandlerChain<Character> chain2 = new WeakHandlerChain<Character>();

        // Create and register automatically...
        new SymbolHandler(chain2);

        // Register the *same* white space handler to different chains!...
        chain2.register(ch);
        // Only handles upper case letters...
        new LetterCaseHandler(chain2, true);
        // Handles *any* type of letter... 
        new LetterHandler(chain2);
        // Only handles lower case letters...
        ch = new LetterCaseHandler(chain2, false);
        
        // Add chain1 as handler for chain2... This includes the sink for logging purposes...
        chain2.register(chain1);

        List<HandlerChain<Character>> chains = new ArrayList<HandlerChain<Character>>(2);
        chains.add(chain1);
        chains.add(chain2);

        out.heading("Created chains and associated handlers");
        
        for (int i = 0; i < chains.size(); i++) {
            HandlerChain<Character> chain = chains.get(i); 
            out.println();
            out.println("Handler chain ", i, ": ", chain);
            for (Handler<Character> h : chain.getHandlers(null)) {
                out.println();
                out.println("Getting chain starting after: ", h, "...");
                List<Handler<Character>> handlers = chain.getHandlers(h);
                for (int j = 0; j < handlers.size(); j++) {
                    Handler<Character> handler = handlers.get(j);
                    out.println("-> ", j, ": ", handler);
                }
            }
            out.println();
        }

        //
        // Test illegal operations
        //
        
        out.heading("Illegal chain operations");
        
        out.println();
        try {
            chain2.register(ch);
            out.error("Registered ", ch, " twice to ", chain2, "!");
            ok = false;
            
        } catch (RuntimeException re) {
            out.println("OK - could not register ", ch, " twice to ", chain2, "!");
            out.println(re);
        }

        out.println();
        try {
            chain1.unregister(ch);
            out.error("Unregistered ", ch, " from chain ", chain1, "!");
            ok = false;
            
        } catch (RuntimeException re) {
            out.println("OK - could not unregister ", ch, " from chain ",  chain1, "!");
            out.println(re);
        }

        out.println();
        try {
            chain1.handle(null);
            out.println("OK - could handle null");
            
        } catch (RuntimeException re) {
            out.error("Could not handle null!");
            out.println(re);
            ok = false;
        }
        
        //
        // Using handlers
        //
        
        out.heading("Using chains");

        CharacterFactory cf = new CharacterFactory();
        Sentence sentence = null; 

        // Use supplied argument, if any, or use default sentence...
        String s = arguments.getArgumentValue("sentence", "- They call me a madman (!!), but I am not mad at anyone...!?");
        // Create...
        sentence = cf.getSentence(s);
        
        out.println();
        out.println("Iterating over characters in: \"", sentence, "\"");
        out.println();
        
        for (int i = 0; i < chains.size(); i++) {
            HandlerChain<Character> chain = chains.get(i); 
            out.println();
            out.println("Handler chain ", i, ": ", chain);

            int n = 0;
            for (Word word : sentence) {
                for (Character c : word) {
                    Handler<?> handler = chain.handle(c);
                    if (handler == null) {
                        out.println("Cannot handle character: ", c);
                    } else {
                        n++;
                    }
                }
            }
            out.println();
            out.println("Total: could handle ", n, "/", sentence.length(), " characters!");
            out.println();
        }

        //
        // Modifying chains
        //
        
        out.heading("Changing chains at execution time");
        
        // Cannot be zero...
        int remove = MathFacade.getInstance().getRandomNumber(1, sentence.length() / 3);
        
        out.println("Removing matching handler every ", remove, " character(s)...");
        
        for (int i = 0; i < chains.size(); i++) {
            HandlerChain<Character> chain = chains.get(i); 
            out.println();
            out.println();
            out.println("Handler chain ", i, ": ", chain);
            
            int j = 0, n = 0;
            for (Word word : sentence) {
                for (Character c : word) {
                    j++;
                    // explicit next link in case the chain cannot handle it...
                    Handler<Character> handler = chain.handle(c, linkToSink); 
                    if (handler == null) {
                        out.println("Cannot handle character: ", c);
                    } else {
                        n++;
                        if (j % remove == remove - 1) {
                            out.warn("Removing handler: ", handler);
                            chain.unregister(handler);
                        }
                    }
                }
            }
            out.println();
            out.println("Total: could handle ", n, "/", j, " characters!");
        }
        
        out.heading("Chain of Responsibility tests ", ok ? "succeeded!" : "failed!");
        
        return ok;
    }
    
}
