/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.chainofresponsibility;

import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;

import dk.rode.thesis.meta.reflect.WeakReferenceEntry;

/**
  * A <i>weak handler chain</i> is a thread-safe handler chain storing 
  * registered {@link Handler handlers} as weak references. <p>
  * 
  * Handler chains can change at runtime, and handlers are internally
  * stored as weak references. Hence, registering a handler to 
  * a chain will not increase its reference count, and they may be
  * garbage collected at any time. The chain automatically 
  * purges handlers that have been garbage collected. <p> 
  * 
  * A handler chain <b>is</b> thread safe. <p>
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  *
  * @see    StandardHandlerChain 
  * 
  * @param  <R> The type of requests handled by this chain.
  */
public class WeakHandlerChain<R> extends AbstractHandlerChain<R> {

    /**
      * The ordered list of handlers, where the first handler
      * is the first inserted element. <p>
      *
      * We use weak references to ensure that when a handler
      * goes out off scope, it does not need to explicitly
      * invoke {@link #unregister(Handler)} (though it may). <p>
      * 
      * <b>Implementation notes</b>:<br>
      * For small chains, a simple list will suffice, but it does 
      * not scale well: if each handler starting from handler 
      * <i>m</i> in the chain reinvokes the chain, we get a 
      * running time of <i>O(n*m)</i>, i.e. worst time <i>O(n*n)</i>, 
      * where <i>n</i> is the chain length. Existence, registration, 
      * and unregistration takes <i>O(n)</i>, but can at the same 
      * time discard garbage collected handlers. <p>
      * 
      * A better solution is to use a linked hash set to avoid 
      * <i>O(n)</i> existence, registration, and unregistration 
      * time, and only purge garbage collected entries when a 
      * list is retrieved. But, since weak references are used,
      * this require a mapping class because weak references
      * use the identity (==) operator to test for equality. 
      * This is the implementation here. <p>
      * 
      * An even faster and more scalable solution would be to use an index 
      * into a list, pointing to the start of the sub-list to return. 
      * A {@link java.util.NavigableMap navigable} map (tree map) can be used 
      * for this, i.e. {@code tailMap(..)} methods, but since we do not enforce 
      * comparable handlers and use weak references, this is actually quite tricky!
      * For this simple example, we are satisfied with our simple solution. <p>
      *
      * Never null.
      */
    private final LinkedHashSet<WeakReferenceEntry<Handler<R>>> handlers;

    /**
      * No-arg constructor.
      */
    public WeakHandlerChain() {
        this.handlers = new LinkedHashSet<WeakReferenceEntry<Handler<R>>>();
    }

    /**
      * {@inheritDoc}
      *
      * This method will purge handlers that have been garbage collected
      * because no longer referenced.
      */
    public synchronized List<Handler<R>> getHandlers(Handler<R> handler) {
        List<Handler<R>> handlers = new LinkedList<Handler<R>>();
        boolean add = (handler == null) || (!this.isRegistered(handler));
        for (Iterator<WeakReferenceEntry<Handler<R>>> i = this.handlers.iterator(); i.hasNext(); ) {
            // Handler no longer in ordinary use? If so, should never be
            // equivalent to "handler" since we are supplying it from the
            // context "now", hence having a reference to it...
            Handler<R> h = i.next().get();
            if (h == null) {
                i.remove();
            } else if (add) {
                handlers.add(h);
            // Do not add until next iteration...
            } else if (h.equals(handler)) {
                add = true;
            }
        }
        return handlers;
    }

    /**
      *
      */
    public boolean isRegistered(Handler<R> handler) {
        return this.handlers.contains(new WeakReferenceEntry<Handler<R>>(handler));
    }

    /**
      * {@inheritDoc}
      *
      * In case contained handlers have been garbage collection since
      * the last call to {@link #getHandlers(Handler)}, the returned
      * size will be larger than the actual size.
      */
    public int size() {
        return this.handlers.size();
    }

    /**
      *
      */
    public synchronized void register(Handler<R> handler) {
        WeakReferenceEntry<Handler<R>> e =
            new WeakReferenceEntry<Handler<R>>(handler);
        if (this.handlers.contains(e)) {
            throw new IllegalArgumentException("Handler already registered");
        }
        this.handlers.add(e);
    }

    /**
      *
      */
    public synchronized void unregister(Handler<R> handler) {
        WeakReferenceEntry<Handler<R>> e =
            new WeakReferenceEntry<Handler<R>>(handler);
        if (!this.handlers.contains(e)) {
            throw new IllegalArgumentException("Handler not registered");
        }
        this.handlers.remove(e);
    }

}
