/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.command;

import static dk.rode.thesis.meta.util.Throwables.toThrowableType;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.log.Log;
import dk.rode.thesis.meta.log.LogFactory;
import dk.rode.thesis.strategy.ObjectPolicy;

/**
  * A <i>command processor</i> can {@link Command#execute() execute}
  * and possibly {@link Command#undo() undo} a given group of 
  * commands. <p> 
  *
  * <b>Implementation notes</b>:<br>
  * This class is a variant of the POSA Command Processor pattern suggested 
  * by Schmidt et al. [<a href="http://www.rode.dk/thesis/Bibliography.htm#Schmidt00" target="thesis-bibliography">Schmidt00</a>, p.397]
  * that also handles <i>commands spawned by other commands</i> during 
  * {@link Command#execute() execution}. This enables commands to express 
  * business-like semantics that can alter the behaviour at runtime depending 
  * on data. For example, a given command could based on data decide to spawn a 
  * new type of command to handle the data invariants, which in turn could spawn 
  * other commands, etc. The execution is depth-first, which ensures that changes 
  * performed by previously spawned commands will be available for later commands 
  * in the (original) list of commands to execute. An example of a command that 
  * possibly will spawn other commands is the {@link ReverseCommand}. <p> 
  *
  * The support for spawned commands lessen the need for  
  * {@link CompositeCommand composite} commands considerably, if not 
  * all together. In the general case, we feel it is better to let the 
  * command processor handle execution and undo of <i>any</i> command, 
  * including those otherwise stored in composites. This ensures uniform 
  * handling of all commands, consistent log format, more control, etc. <p>  
  * 
  * The supplied group of commands acts as part of a local <i>history list</i>,
  * though in conjunction with possibly spawned commands, as described by 
  * Gamma et al. [<a href="http://www.rode.dk/thesis/Bibliography.htm#Gamma95" target="thesis-bibliography">Gamma95</a>, p.238], 
  * but processors are stateless otherwise. <p>
  * 
  * Command processors are not generic, but use generic methods to handle 
  * any type of command result value. 
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @see    #execute(List)
  * @see    CommandProcessingResult
  */
@Participant("Invoker")
public class CommandProcessor {
    
    /**
      * No-arg constructor.
      */
    public CommandProcessor() {
        super();
    }

    /**
      * {@link Command#execute() Executes} the commands supplied as 
      * {@code commands}, in order, or tries to {@link Command#undo() 
      * undo} executed commands in case of execution failure. <p>
      *
      * If the execution and undo fails, an exception is thrown. Otherwise
      * a {@link CommandProcessingResult} is returned with the status and
      * possible result of the execution. The result is that of the last
      * executed command, which may be a spawned command.
      * 
      * @param  <E>      The type of result obtained by executing the commands
      *                  in {@code commands}.
      * @param  commands The group of commands to execute; cannot be null. 
      *                  A null command will cause an execution failure with a 
      *                  following attempt at undoing of previous executed commands.
      * @return The result of the execution; never null.
      * @throws NullPointerException If {@code commands} is null.
      * @throws CommandException If the execution (and undo) fails.
      * @see    #execute(List)                           
      */
    public <E> CommandProcessingResult<E> execute(Command<E>... commands) throws CommandException {
        return this.execute(Arrays.asList(commands));
    }
    
    /**
      * {@link Command#execute() Executes} the commands supplied as 
      * {@code commands}, in order, or tries to {@link Command#undo() 
      * undo} executed commands in case of execution failure. <p>
      *
      * If the execution and undo fails, an exception is thrown. Otherwise
      * a {@link CommandProcessingResult} is returned with the status and
      * possible result of the execution. The result is that of the last
      * executed command, which may be a spawned command.
      * 
      * @param  <E>      The type of result obtained by executing the commands
      *                  in {@code commands}.
      * @param  commands The list (group) of commands to execute. A null 
      *                  entry will cause an execution failure with a 
      *                  following attempt at undoing of previous executed 
      *                  commands.
      * @return The result of the execution; never null.
      * @throws NullPointerException If {@code commands} is null.          
      * @throws CommandException If the execution (and undo) fails.
      * @see    #execute(Command...)                           
      */
    public <E> CommandProcessingResult<E> execute(List<Command<E>> commands) throws CommandException {
        // The "executedCommands" list is a history list for undo. We cannot simply 
        // use "commands" because each command may spawn additional commands that 
        // must be undone as well, if possible...
        List<Command<E>> executedCommands = new ArrayList<Command<E>>((int)(commands.size() * 1.2));
        // Will also be true if "commands" is empty...
        boolean executed = this.execute(commands, executedCommands);
        // If not executed, the result is always null...
        E result = null;
        if ((executed) && (executedCommands.size() > 0)) {
            result = executedCommands.get(executedCommands.size() - 1).getResult(); 
        }
        return new CommandProcessingResult<E>(executed, result);
    }
        
    /**
      * {@link Command#execute() Executes} the commands supplied as 
      * {@code commands}, in order, or tries to {@link Command#undo() 
      * undo} executed commands in case of execution failure. <p>
      *
      * Each executed command is added to the history list of executed
      * commands supplied as {@code executedCommands}, which is the
      * list that will be tried {@link #undo(List) undone} in case
      * of execution failure. <p>
      * 
      * If the execution and undo fails, an exception is thrown. If the
      * execution fails, but all commands could be undone, this method
      * returns false. Otherwise true to indicate execution success.
      * 
      * @param  <E>              The type of result obtained by executing the commands
      *                          in {@code commands}.
      * @param  commands         The list (group) of commands to execute. A null 
      *                          entry will cause an execution failure with a 
      *                          following attempt at undoing of previous executed 
      *                          commands.
      * @param  executedCommands The history list of executed commands; never null.                  
      * @return True if the execution succeeded, false if it failed but could
      *         be undone.
      * @throws NullPointerException If either argument is null.          
      * @throws CommandException If the execution and undo fails.
      */
    private <E> boolean execute(List<Command<E>> commands, List<Command<E>> executedCommands) throws CommandException {
        if (commands == null) {
            throw new NullPointerException("Commands cannot be null");
        } 
        int index = 1, length = commands.size();
        // No commands supplied? 
        if (length == 0) {
            return true;
        }
        
        // Current command...
        Command<E> command = null;
        Log out = LogFactory.getLog(this);
        try {
            // Execute in order...
            for (int i = 0; i < length; i++, index++) {
                command = commands.get(i);
                // Add before execution, so it can be undone...
                executedCommands.add(command);
                
                // Execute...
                
                // "command" may be null, which will cause the execution
                // of the previous executed commands to be undone...
                List<Command<E>> spawned = command.execute();
                out.println("Executed ", index, "/", length, ": ", command, " -> ", command.getResult());
                
                // Can it be undone?...
                if (!command.isUndoable()) {
                    out.warn("Command is not undoable after execution: ", command);
                }
                
                // Recurse for spawned commands, if any...
                if ((spawned != null) && (spawned.size() > 0)) {
                    out.println("Executing ", spawned.size(), " spawned command(s)...");
                    
                    // If the execution failed but could be undone while recursing, 
                    // we have to stop as well...
                    try {
                        if (!this.execute(spawned, executedCommands)) {
                            return false;
                        }
                    } finally {
                        out.println("Executing ", spawned.size(), " spawned command(s) done!");
                    }
                }
            }
            
            // All executed successfully...
            return true;
            
        // In case the execution of "command"
        } catch (Exception e) {
            out.error("Command execution failed ", index, "/", length, ": ", command).error(e);

            // Try and undo all executed commands, even undoable ones... The 
            // return value from undo determines if the undo was successful or not,
            // but it may also generate an exception...
            boolean undone = false;
            try {
                undone = this.undo(executedCommands);
                
            } catch (Exception ee) {
                // Fail-through...
            }
            // Re-throw original exception as a proper type if we could not
            // undo the commands...
            if (!undone) {
                throw toThrowableType(e, CommandException.class);
            }
            
            // Execution failed, but all could be undone...
            return false;
        }
    }

    /**
      * Undoes the executed {@link Command commands} supplied as 
      * {@code executedCommands}, if possible. <p>
      *   
      * If the undo of a given {@link Command#isUndoable() undoable} command 
      * fails, this method will throw an exception. If a given command is
      * not undoable, this method returns false. True is returned if all 
      * commands could be undone.    
      *  
      * @param  <E>              The type of result obtained from the executed 
      *                          commands.
      * @param  executedCommands The executed commands to undo; cannot be null.
      * @return True if all commands could be undone, false if not.
      * @throws NullPointerException If {@code executedCommands} is null.
      * @throws Exception If the undo of a command fails.
      */
    private <E> boolean undo(List<Command<E>> executedCommands) throws Exception {
        Log out = LogFactory.getLog(this);
        Exception exception = null;
        boolean undone = true;
        int index = executedCommands.size(), length = index;
        // Undo in reverse...
        for (int i = index - 1; i >= 0; i--, index--) {
            Command<E> command = executedCommands.get(i);
            out.println("Undoing ", index, "/", length, ": ", command);
            // In case the fault was due to a null pointer...
            if (command == null) {
                continue;
            }
            
            // Is command undoable?
            if (command.isUndoable()) {
                try {
                    command.undo();
                    
                } catch (Exception e) {
                    exception = e;
                    out.error("Undo failed: ", command).error(e);
                    undone = false;
                    break;
                }
                
            } else {
                out.warn("Command is not undoable: ", command);
                undone = false;
                break;
            }
        }
        if (undone) {
            out.println("All commands successfully undone");
        } else {
            out.warn("Stopping undo!");
            // Re-throw...
            if (exception != null) {
                throw exception;
            }
        }
        return undone;
    }
    
    /**
      * Returns the string representation of this processor.
      * 
      * @return The string representation; never null.
      */
    public String toString() {
        return ObjectPolicy.ID.asString(this);
    }
    
}
