/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 * 
 */
package dk.rode.thesis.facade;

import java.math.BigInteger;

import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.model.AbstractSequence;
import dk.rode.thesis.meta.model.Sequence;

/**
  * A <i>Fibonacci sequence</i> represents an integer sequence, where
  * each number delivered by {@link #next()} is the sum of the two 
  * preceding numbers, except one which is the initial value: <p>
  * 
  * 1, 1, 2, 3, 5, 8, 13, 21, 34, 55, 89, 144, 233, 377, 610, 987, 1597, .. <p>
  *  
  * A Fibonacci sequence is {@link Sequence#bounded() unbounded} unless
  * a maximum number to consider has been supplied at construction time, not
  * {@link #unique() unique} (1), but {@link Sequence#consistent() consistent}.
  *  
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  */
@Participant("SubsystemClass")
public class FibonacciSequence extends AbstractSequence<BigInteger> implements Sequence<BigInteger> {

    /**
      * The previous Fibonacci number.  
      */
    private BigInteger previous;
    /**
      * The current Fibonacci number. <p>
      * 
      * The first number is one.
      */
    private BigInteger current;
    /**
      * Maximum number to consider. <p>
      * 
      * The number need not be a Fibonacci number. <p>
      * 
      * A value of null means no maximum.
      */
    private BigInteger maximum;
    
    /**
      * No-arg constructor. <p>
      * 
      * No maximum number is enforced.
      */
    public FibonacciSequence() {
        this((BigInteger)null);
    }
    
    /**
      * Constructor.
      * 
      * The maximum Fibonacci number to be delivered will the smallest
      * Fibonacci number smaller or equal to {@code maximum}, if not
      * null.
      * 
      * @param  maximum The maximum bound; can be null.
      * @throws IllegalArgumentException If {@code maximum} is not null, but
      *         not positive.
      *       
      */
    public FibonacciSequence(BigInteger maximum) {
        super(true);
        if ((this.maximum = maximum) != null) {
            if (this.maximum.signum() != 1) {
                throw new IllegalArgumentException("Maximum must be positive");
            }
        }
    }
    
    /**
      * Copy constructor.
      * 
      * @param  sequence The sequence to copy; cannot be null.
      * @throws NullPointerException If {@code sequence} is null.
      */
    public FibonacciSequence(FibonacciSequence sequence) {
        super(sequence);
        this.previous = sequence.previous;
        this.current = sequence.current;
        this.maximum = sequence.maximum;
    }
    
    /**
      * 
      */
    public FibonacciSequence copy() {
        return new FibonacciSequence(this); 
    }
    
    /**
      * 
      */
    public boolean bounded() {
        return (this.maximum != null);
    }

    /**
      * Returns true.
      *  
      * @return True.
      */
    public boolean consistent() {
        return true;
    }

    /**
      * Returns false.
      *  
      * @return False.
      */
    public boolean unique() {
        return false; // 1...
    }
    
    /**
      * 
      */
    public BigInteger current() {
        return this.current;
    }

    /**
      * 
      */
    public BigInteger next() {
        if (this.state != State.NORMAL) {
            this.state = State.NORMAL;
        }
        BigInteger previous = this.previous;
        this.previous = this.current;
        this.current = this.current.add(previous);
        if ((this.maximum != null) && (this.current.compareTo(this.maximum) > 0)) {
            this.reset();
            this.state = State.RESTART;
        }
        return this.current;
    }

    /**
      * 
      */
    public void reset() {
        super.reset();
        this.previous = BigInteger.ZERO;
        this.current = BigInteger.ONE;
    }

}
