/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.factorymethod;

import dk.rode.thesis.command.Command;
import dk.rode.thesis.command.NullCommand;
import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.log.Log;
import dk.rode.thesis.meta.log.LogFactory;
import dk.rode.thesis.meta.model.Sequence;
import dk.rode.thesis.strategy.ObjectPolicy;

/**
  * A <i>command creator</i> creates {@link Command commands} to 
  * manipulate a given receiver {@link Sequence sequence} using the 
  * {@link #getCommand(Sequence, Object)} method. <p>
  * 
  * The type of command to create is determined by a <i>token</i> supplied 
  * as the last argument to the {@code getCommand(..)} method. The type of 
  * token is determined by the type parameter {@code T}. <p>
  * 
  * Sub-classes <b>must</b> implement the {@link #create(Sequence, Object)}
  * factory method, and <b>can</b> override the {@link #createDefault(Sequence)} 
  * factory method to change the default behaviour. The {@link NullCommand}
  * is returned as the default command by this implementation, which is
  * thus a command without a sequence as a receiver. <p>
  * 
  * A command creator is stateless and thread-safe. <p>
  * 
  * <b>Implementation notes</b>:<br>
  * This class is declared abstract to allow for the factory methods to be 
  * protected. If it had been defined as an interface, the factory methods 
  * would have had to been declared public. <p>
  * 
  * The design requires that type parameter, {@code E}, of the created 
  * commands must be supplied as a type parameter to the creator class and
  * not via a generic method, which is a viable alternative. This is to allow 
  * sub-classes to utilise the specific type of {@code E} in conjunction 
  * with the type of tokens used, i.e. the tokens may also depend on 
  * {@code E}. Such a dependency is not possible using generic methods
  * alone. The {@link ReflectiveCommandCreator} class express such a
  * dependency between {@code E} and {@code T} because {@code T} uses
  * {@code E} as a generic type parameter. <p>
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @param  <E> The type of values delivered by created commands (and thus sequences).
  * @param  <T> The type of token used to identify the type of {@code Command}
  *             to create.
  */
@Participant("Creator")
public abstract class CommandCreator<E,T> {

    /**
      * No-arg constructor.
      */
    protected CommandCreator() {
        super();
    }

    /**
      * Returns a new command instance of the type identified by {@code token}
      * associated with the {@link Sequence sequence} receiver supplied 
      * at {@code sequence}. <p>
      * 
      * The creation is deferred to the {@link #create(Sequence, Object)}
      * factory method, but if the creation fails, new <i>default command type</i> 
      * is tried created with the {@link #createDefault(Sequence)} method. If 
      * that fails as well, a {@link NullCommand} is returned. <p>
      * 
      * If {@code token} is null, a null command is returned without
      * deferring a call to the create methods. <p>
      * 
      * If a token cannot be handled, a default command is returned. 
      * 
      * @param  sequence The sequence receiver; can be null, in which
      *                  case a default (or null) command is returned.
      * @param  token    The token specifying the type of command; can
      *                  be null, in which case a null command is returned.
      * @return A new command; never null.
      * @see    #create(Sequence, Object)
      * @see    #createDefault(Sequence) 
      */
    public final Command<E> getCommand(Sequence<E> sequence, T token) {
        Log out = LogFactory.getLog(this);
        Command<E> command = null;        
        if (sequence == null) {
            command = new NullCommand<E>();
            out.println("Created command (null sequence): ", command);
            return command;
        }
        // Create...
        try {
            if ((command = this.create(sequence, token)) != null) {
                out.println("Created command: ", token, " -> ", command);
            } else {
                out.warn("Cannot handle token: ", token);
            }
            
        } catch (Exception e) {
            out.warn("Command creation failed: ", token, " -> ", sequence).warn(e);
            // Fall-through...
        }
        
        // Still no command?...
        if (command == null) {
            // Try default...
            try {
                if ((command = this.createDefault(sequence)) != null) {
                    out.println("Created default command: ", token, " -> ", command);
                }
            } catch (Exception e) {
                out.error("Default command creation failed: ", token, " -> ", sequence).error(e);
                // Fall-through...
            }
            // If even the default creation failed, create and return
            // a null command so we never return null...
            if (command == null) {
                command = new NullCommand<E>();
                out.println("Created fail-safe default command: ", token, " -> ", command);
            }
        }
        return command;
    }
    
    /**
      * Factory method to create a new {@link Command command} of the
      * type specified by the <i>token</i> supplied as {@code token}. <p>
      *  
      * The {@link Sequence sequence} set as the receiver is supplied
      * as {@code sequence}.
      *    
      * @param  sequence The sequence; never null.
      * @param  token    The token identifying the type of command to create;
      *                  can be null.
      * @return The corresponding command, or null if no such command
      *         could be created. If null is returned, a default (on null)
      *         command is returned by {@link #getCommand(Sequence, Object)}.
      * @throws Exception If the creation fails; if so, a default (or null)
      *         command will be returned by {@code getCommand}.
      * @see    #createDefault(Sequence)              
      */
    protected abstract Command<E> create(Sequence<E> sequence, T token) throws Exception;
    
    /**
      * Factory method to create a new default command in case the
      * {@link #create(Sequence, Object) normal} command creation 
      * failed. <p>
      * 
      * @param  sequence The sequence; never null.
      * @return A new default command; if null is returned, a null command
      *         is returned by {@link #getCommand(Sequence, Object)}.
      *         This implementation returns a new {@link NullCommand}. 
      * @throws Exception If the creation fails; if so, a null command
      *         will be returned by {@code getCommand}.     
      */
    protected Command<E> createDefault(Sequence<E> sequence) throws Exception {
        return new NullCommand<E>();
    }
    
    /**
      * Returns the hash code of this command creator. 
      *
      * @return The hash code.
      */
    public int hashCode() {
        return this.getClass().hashCode();
    }
    
    /**
      * Returns true if class of {@code object} is equal to
      * this class, false if not.
      *
      * @param  object The object to test; can be null.
      * @return True if equal, false if not.
      */
    public boolean equals(Object object) {
        if (object == null) {
            return false;
        } else if (object == this) {
            return true;
        } else if (this.getClass().equals(object.getClass())) {
            return true;
        }
        return false;
    }
    
    /**
      * Returns the string representation of this command creator.
      *  
      * @return The string representation; never null.
      */
    public String toString() {
        return ObjectPolicy.IDENTITY.asString(this);
    }
    
}
