/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.factorymethod;

import java.util.Arrays;

import dk.rode.thesis.abstractfactory.PrototypicalFactory;
import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.reflect.InstantiableTypeLiteral;
import dk.rode.thesis.strategy.ObjectPolicy;

/**
  * A <i>generic factory</i> that will create products of type {@code T}. <p>
  * 
  * Usage: to create instances of {@code Foo<String>} using the constructor
  * accepting a {@code String} and {@code int} argument, in that order: 
  * <pre>
  *   // Create factory...
  *   Factory&lt;Foo&lt;String>> factory = 
  *     new {@link #Factory(Class...) Factory}(String.class, Integer.class){}; // anonymous class
  *     
  *   // Create products...  
  *   Foo&lt;String> foo    = factory.{@link #newInstance(Object...) newInstance}("Bar", 42);
  *   Foo&lt;String> foobar = factory.{@link #newInstance(Object...) newInstance}("FooBar", 117);
  * </pre>
  *
  * The constructor to create new instances of {@code T} is
  * fetched ignoring primitive types, e.g. a formal parameter type
  * of {@code Integer} will match both {@code Integer} and {@code int}. <p>
  * 
  * <b>Implementation notes</b>:<br>
  * This factory is classified as a Factory Method, not an 
  * {@linkplain dk.rode.thesis.abstractfactory Abstract Factory},
  * because its traits most closely resemble those described for Factory
  * Method by Gamma et al. [<a href="http://www.rode.dk/thesis/Bibliography.htm#Gamma95" target="thesis-bibliography">Gamma95</a>, p.107-116], but it is a judgement call. 
  * We see it as a variant of the <i>Creator</i> class [<a href="http://www.rode.dk/thesis/Bibliography.htm#Gamma95" target="thesis-bibliography">Gamma95</a>, p.113], which 
  * uses templates to determine the actual type of product to create. We still 
  * use it in the Abstract Factory implementation as well, though. These are
  * the reasons why it is named {@code Factory}, and not {@code Creator}. <p>
  * 
  * The {@link PrototypicalFactory} shows how the 
  * {@linkplain dk.rode.thesis.prototype Prototype} pattern can 
  * often be used as an alternative to reflection. <p>
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @see    InstantiableTypeLiteral
  * @see    TypedFactory
  * 
  * @param  <T> The product type created by this factory.
  */
@Participant("Creator")
public abstract class Factory<T> extends InstantiableTypeLiteral<T> {

    /**
      * Constructor, which creates this factory to create instances
      * of the type represented by the type parameter {@code T} using
      * the declared constructor having the formal parameter 
      * types supplied as {@code parameterTypes}, if any. <p>
      * 
      * The constructor is made accessible if not already. New instances
      * of {@code T} must be created with the {@link #newInstance(Object...)}
      * factory method.
      *        
      * @param  parameterTypes The formal parameter types for the 
      *                        declared constructor to use.
      * @throws IllegalArgumentException If this class is not generic, if
      *         {@code T} has no raw type, or if the raw raw represents 
      *         a non instantiable class.
      * @throws NoSuchMethodException If no such constructor is declared in 
      *         the raw type of {@code T}. 
      */
    protected Factory(Class<?>... parameterTypes) throws NoSuchMethodException {
        super(parameterTypes);
    }

    /**
      * Returns the string representation of this factory.
      * 
      * @return The string representation; never null.
      */
    public String toString() {
        StringBuilder sb = (StringBuilder)ObjectPolicy.NAME_OPEN.toString(this);
        sb.append("type: ");
        if (this.type instanceof Class) {
            sb.append(((Class<?>)this.type).getName());
        } else {
            sb.append(this.type);
        }
        return sb.append("; constructor: ")
                 .append(Arrays.asList(this.getConstructor().getParameterTypes()))
                 .append(']').toString();
    }

}
