/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.flyweight;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.strategy.ObjectPolicy;

/**
  * A <i>character factory</i> creates and manages {@link Character}, 
  * {@link Word}, and {@link Sentence} objects. The first two 
  * types are flyweights, while the latter is an unshared 
  * flyweight. <p>
  * 
  * This factory supports an invariant that states that any 
  * {@link Letter} can only contain a single char. Sub-classes 
  * may override it to alter this behaviour. <p>
  * 
  * Character factories are thread-safe. <p>
  * 
  * <b>Implementation notes</b>:<br>
  * The {@link #getWords()} and {@link #getCharacters()} method return
  * unmodifiable collections. This corresponds to <i>protection proxies</i>
  * as described in the {@linkplain dk.rode.thesis.proxy Proxy} pattern.
  *
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  */
@Participant("FlyweightFactory")
public class CharacterFactory {

    /**
      * The set of created characters so far. <p>
      * 
      * Never null, but may be empty.
      */
    private final Map<String,Character> characters;
    /**
      * The set of created words so far. <p>
      * 
      * Never null, but may be empty.
      */
    private final Map<Word,Word> words; // Map because there is no way to "get" from a set without iteration
    
    /**
      * No-arg constructor.
      */
    public CharacterFactory() {
        this.characters = new HashMap<String,Character>();
        this.words = new HashMap<Word,Word>();
    }
    
    /**
      * Returns a <b>read-only</b> collection of the words created 
      * by this factory.
      *  
      * @return A collections of created words; never null, but may
      *         be empty.
      * @see    #getCharacters()          
      */
    public Collection<Word> getWords() {
        return Collections.unmodifiableSet(this.words.keySet()); 
    }
    
    /**
      * Returns a <b>read-only</b> collection of the characters created 
      * by this factory.
      *  
      * @return A collection of created characters; never null, but may
      *         be empty.
      * @see    #getWords          
      */
    public Collection<Character> getCharacters() {
        return Collections.unmodifiableCollection(this.characters.values());
    }

    /**
      * Returns the created characters of the type supplied
      * as {@code type}, if ant.
      * 
      * @param  <C>  The type of character.
      * @param  type The character type; cannot be null.
      * @return A list containing the created characters by this
      *         factory; never null, but can be empty. 
      * @throws NullPointerException If {@code type} is null.
      * @see    #count(Class)
      */
    public final synchronized <C extends Character> List<C> getCharacters(Class<C> type) {
        // Synchronised, or we could get a concurrent modification exception...
        List<C> list = new ArrayList<C>();
        for (Character c : this.characters.values()) {
            if (type.isInstance(c)) {
                list.add(type.cast(c));
            }
        }
        return list;
    }

    /**
      * Returns the number of instances created by this 
      * factory of the specific {@link Character} type
      * supplied as {@code type}.
      * 
      * @param  <C>  The type of character.
      * @param  type The character type; cannot be null.
      * @return The count. 
      * @throws NullPointerException If {@code type} is null.
      * @see    #getCharacters(Class)
      */
    public final synchronized <C extends Character> int count(Class<C> type) {
        // Synchronised, or we could get a concurrent modification exception...
        int count = 0;
        for (Character c : this.characters.values()) {
            if (type.isInstance(c)) {
                count++;
            }
        }
        return count;
    }
    
    /**
      * Returns the number of created words by this factory.
      * 
      * @return The number of words.
      */
    public final int countWords() {
        return this.words.size();
    }
    
    /**
      * Returns the unique {@link Character} instance representing
      * the char value supplied as {@code character}. <p>
      * 
      * @param  character The char. 
      * @return The corresponding {@code Character} instance; never null.
      */
    public final Character getCharacter(char character) {
        return this.getCharacter(java.lang.Character.toString(character));
    }
    
    /**
      * Returns the unique {@link Character} instance representing
      * the char value at index {@code index} in {@code character}. 
      * 
      * @param  character The char sequence supplying the desired char
      *                   at {@code index}; cannot be null.
      * @param  index     The index to use.
      * @return The corresponding {@code Character} instance; never null.
      * @throws NullPointerException If {@code character} is null.
      * @throws IndexOutOfBoundsException If {@code index} is illegal.
      */
    public final Character getCharacter(CharSequence character, int index) {
        return this.getCharacter(character.subSequence(index, index + 1).toString());
    }

    /**
      * Returns the unique {@link Character} instance representing
      * the value supplied as {@code character}. <p>
      * 
      * @param  character The char. 
      * @return The corresponding {@code Character} instance; never null.
      * @throws NullPointerException If {@code character} is null.
      * @throws IllegalArgumentException If {@code character} does not 
      *         contain a single char.
      */
    public final synchronized Character getCharacter(String character) {
        Character c = this.characters.get(character = this.normalise(character));
        if (c == null) {
            this.characters.put(character, c = this.create(character)); 
        }
        return c;
    }

    /**
      * Possibly creates and returns the {@link Word} corresponding to
      * the characters supplied as {@code word}.
      * 
      * @param  word The characters forming the word, in order; never null.
      * @return The corresponding {@code Word} instance; never null.
      */
    private final synchronized Word getWord(List<Character> word) {
        Character[] characters = new Character[word.size()];
        Word w = new Word(word.toArray(characters));
        if (this.words.containsKey(w)) {
            w = this.words.get(w);
        } else {
            this.words.put(w, w);
        }
        return w;
    }

    /**
      * Returns a {@link Word} instance parsed from the letters in string. <p>
      * 
      * If more than one word is present in {@code word}, the first one
      * is returned. 
      * 
      * @param  word The word as a string; cannot be null.
      * @return A corresponding {@code Word} instance; never null.
      * @throws NullPointerException If {@code word} is null.
      */
    public final Word getWord(String word) {
        for (Word w : this.getSentence(word)) {
            return w;
        }
        // Will never happen...
        throw new IllegalStateException("Could not get word from: " + word);
    }
    
    /**
      * Returns a new {@link Sentence} containing the words parsed from
      * {@code string}.
      * 
      * @param  sentence The actual sentence as a string; cannot be null.
      * @param  start    The start index.
      * @param  end      The end index.
      * @return A sentence representing {@code string}; can be null if
      *         {@code string} only contains white space characters.
      * @throws NullPointerException If {@code sentence} is null.
      * @throws IndexOutOfBoundsException If {@code start} or {@code end}
      *         are illegal.
      */
    public final Sentence getSentence(CharSequence sentence, int start, int end) {
        return this.getSentence(sentence.subSequence(start, end).toString());
    }

    /**
      * Returns a new {@link Sentence} containing the words parsed from
      * {@code sentence}.
      * 
      * @param  sentence The actual sentence as a string; cannot be null.
      * @return A sentence representing {@code sentence}; never null.
      * @throws NullPointerException If {@code sentence} is null.         
      */
    public Sentence getSentence(String sentence) {
        // *Very* simple parser...
        
        // Invariant: string never starts or ends with white space.
        sentence = sentence.trim();
        
        // The sentence...
        List<Word> s = new ArrayList<Word>();
        // The current word...
        List<Character> word = new ArrayList<Character>();
        
        Word last = null;
        
        // Go through each character and build the proper words.
        // Non letters are the delimiter used...
        for (char c : sentence.toCharArray()) {
           Character character = this.getCharacter(c);

           // A word ends on:
           // - whitespace
           // - symbol
           // Both whitespace and symbol double as their own words

           // Continue as usual...
           if (character instanceof Letter) {
               word.add(character);
               continue;
           }
           // Add word...
           if (word.size() > 0) {
               s.add(last = this.getWord(word));
               // Prepare for white space or symbol...
               word.clear();
           }
           word.add(character);
           
           Word current = this.getWord(word);           
           // Skip multiple white space...
           if ((current.isWhiteSpace()) && (current.equals(last))) {
               word.clear();
               continue;
           }
           // Add word...
           s.add(last = current);
           // Prepare for next word...
           word.clear();
        }
        
        // Remember trailing characters, which the invariant says
        // can never be white space...
        if (word.size() > 0) {
            s.add(this.getWord(word));
            
        // All white space? Add a single space so we never return null...
        } else if (s.size() == 0) {
            // Generate white space word once and for all...
            word.add(this.getCharacter(Whitespace.space));
            s.add(this.getWord(word));
        }
        
        // Now create and return the sentence as required...
        Word[] words = new Word[s.size()];
        return new Sentence(s.toArray(words));
    }
    
    /**
      * Normalises the string representing a {@link Character} to
      * a format supported by this factory. <p>
      * 
      * Sub-classes can override this hook for additional functionality.
      * 
      * @param  character The character string; cannot be null.
      * @return The normalised string; never null.
      * @throws NullPointerException If {@code character} is null.
      * @throws IllegalArgumentException If {@code character} is empty, or 
      *         if after trimming contains more than one character.
      * @see    #create(String)          
      */
    protected String normalise(String character) {
        if (character == null) {
            throw new NullPointerException("Character cannot be null");
        }
        int length = character.length();
        if (length == 0) {
            throw new IllegalArgumentException("Character cannot be empty");
            
        } else if (length > 1) {
            String s = character.trim();
            // All whitespace...
            if (s.length() == 0) {
                return Whitespace.space;
                
            } else if (s.length() > 1) {
                throw new IllegalArgumentException("Character cannot be more than one char");
            } else {
                return s;
            }
            
        // All white space represented equally...
        } else if (java.lang.Character.isWhitespace(character.charAt(0))) {
            return Whitespace.space;
        }
        return character;
    }
    
    /**
      * Factory method that creates a {@link Character} based on
      * the character string supplied as {@code character}. <p>
      * 
      * Sub-classes can override this hook for additional functionality.
      * 
      * @param  character The normalised character string; never null.
      * @return The corresponding {@code Character} instance; never null.
      * @see    #normalise(String)
      */
    protected Character create(String character) {
        // Invariant: default normalisation method supplies a string
        //            with a single char!
        
        char c = character.charAt(0);
        
        // Whitespace: will only be created once for this factory!
        if (java.lang.Character.isWhitespace(c)) {
            return new Whitespace();
            
        // Letter...
        } else if (java.lang.Character.isLetter(c)) {
            return new Letter(c);
            
        // Any other symbol, including digits...
        } else {
            return new Symbol(c);
        }
    }
    
    /**
      * Return the string representation of this factory.
      * 
      * @return The string representation; never null.
      */
    public String toString() {
        StringBuilder sb = (StringBuilder)ObjectPolicy.IDENTITY_OPEN.toString(this);
        return sb.append("; characters: ") 
                 .append(this.characters.size())
                 .append("; words: ")
                 .append(this.words.size())
                 .append(']').toString();
    }
    
}
