/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.flyweight;

import dk.rode.thesis.strategy.ObjectPolicy;
import dk.rode.thesis.strategy.StringablePolicy;

/**
  * Default policies for formatting {@link Character} objects
  * into char sequences (<b>not</b> part of the core Flyweight 
  * implementation). <p>
  * 
  * <b>Implementation notes:</b><br>
  * <b>This is NOT part of the core Flyweight functionality, merely 
  * to provide better log descriptions!</b><p>
  * 
  * See implementation notes for the {@link ObjectPolicy} enumeration.
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  */
public enum CharacterPolicy implements StringablePolicy<Character> {

    /**
      * Generates a char sequence representation of a {@link Character} 
      * having the following format.
      * <pre>
      *   Character
      * </pre>
      * 
      * Where {@code Character} is the simple class name of the class
      * of the character. <p>
      * 
      * This policy is considered {@link dk.rode.thesis.strategy.StringablePolicy.Type#CLOSED}. 
      */
    NAME { 
        public Type type() {
            return Type.CLOSED;
        }
    },

    /**
      * Generates a char sequence representation of a {@link Character} 
      * having the following format.
      * <pre>
      *   ..
      * </pre>
      * 
      * Where {@code ..} is the value of the character, i.e. its
      * string representation. <p>
      * 
      * This policy is considered {@link dk.rode.thesis.strategy.StringablePolicy.Type#CLOSED}. 
      */
    VALUE {
        public CharSequence toString(Character character) {
            return character.toString();
        }
        public Type type() {
            return Type.CLOSED;
        }
    },

    /**
      * Generates a char sequence representation of a {@link Character} 
      * having the following format.
      * <pre>
      *   ..
      * </pre>
      * 
      * Where {@code ..} is the upper-cased value of the character, i.e. its
      * string representation upper-cased. <p>
      * 
      * This policy is considered {@link dk.rode.thesis.strategy.StringablePolicy.Type#CLOSED}. 
      */
    VALUE_UPPERCASE {
        public CharSequence toString(Character character) {
            return character.toString().toUpperCase();
        }
        public Type type() {
            return Type.CLOSED;
        }
    },
    
    /**
      * Generates a char sequence representation of a {@link Character} 
      * having the following format.
      * <pre>
      *   Character[..]
      * </pre>
      * 
      * Where {@code Character} is the simple class name of the class
      * of the character, and {@code ..} is the value of the character, 
      * respectively.
      */
    SHORT {
        public CharSequence toString(Character character) {
            StringBuilder sb = (StringBuilder)super.toString(character);
            return sb.append('[').append(character.toString()).append(']');
        }
    },
    
    /**
      * Generates a char sequence representation of a {@link Character} 
      * having the following format.
      * <pre>
      *   Character[hash-code: ..; value: ".."]
      * </pre>
      * 
      * Where {@code Character} is the simple class name of the class
      * of the character, {@code ..} is the hash-code, and
      * {@code ".."} value of the character (excluding quotation marks), 
      * respectively.
      */
    IDENTITY {
        public CharSequence toString(Character character) {
            StringBuilder sb = (StringBuilder)ObjectPolicy.IDENTITY_OPEN.toString(character);
            return sb.append("; value: \"")
                     .append(character.toString())
                     .append("\"]");
        }
    };
    
    /**
      * 
      */
    public CharSequence toString(Character character) {
        return ObjectPolicy.NAME.toString(character);
    }

    /**
      * @return {@link dk.rode.thesis.strategy.StringablePolicy.Type#APPENDABLE} 
      *         is returned unless explicitly documented otherwise for a given constant. 
      */
    public Type type() {
        return Type.APPENDABLE;
    }
        
}
