/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.interpreter;

import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.log.LogFactory;

/**
  * A <i>constant expression</i> represents an expression that
  * can be assigned a constant value. <p>
  * 
  * Constants can be stored in a {@link Context context} as well,
  * but this is not a requirement. <p>
  * 
  * This class does <b>not</b> declare a copy constructor, and will when
  * {@link #copy() copied} return the same instance since the
  * value is constant. <p>
  *
  * <b>Implementation notes</b>:<br>
  * The auto-generated {@link VariableExpression#name name} of a 
  * constant ensures that two constants of the same 
  * {@link VariableExpression#type type}, {@code E}, having the 
  * same actual value will also have the same name. Hence, look-up 
  * of different constant instances in a {@code Context} will 
  * result in a match if same type and value; {@code equals(Object)}
  * works in a similar manner. However, this also means that a constant 
  * cannot be assigned a value during evaluation, not even once, 
  * because the name depends on the value. Trying to create an 
  * {@link AssignmentExpression assignment} expression using a constant 
  * as the target of the evaluation is an error, and will cause an 
  * exception to be thrown at construction time. <p> 
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @param  <E> The type of value the evaluation of this expression 
  *             produces.
  */
@Participant("NonTerminalExpression")
public class ConstantExpression<E> extends VariableExpression<E> {

    /**
      * The value of this constant. <p>
      * 
      * Never null.
      */
    public final E value;

    /**
      * Constructor. <p>
      * 
      * The symbolic representation of this variable will not 
      * include type information regarding {@code type}.
      * 
      * @param  type  The type of the value this expression represents;
      *               cannot be null.
      * @param  value The constant value; cannot be null.
      * @throws NullPointerException If either argument is null.
      */
    public ConstantExpression(Class<E> type, E value) {
        this(type, value, false);
    }
    
    /**
      * Constructor. <p>
      * 
      * The symbolic representation of this constant will 
      * include type information regarding {@code type}
      * if {@code includeTypeInfo} is true, not if false.
      * 
      * @param  type            The type of the value this expression represents;
      *                         cannot be null.
      * @param  value           The constant value; cannot be null.
      * @param  includeTypeInfo True to include type information in
      *                         the symbolic representation, false if not.             
      * @throws NullPointerException If either argument is null.
      */
    public ConstantExpression(Class<E> type, E value, boolean includeTypeInfo) {
        // Ensures inherited "hashCode" and "equals" work. By including
        // the value before the type, we get better hash codes and
        // faster comparison using equals...
        super(type, value.toString() + "@" + type.getName(), includeTypeInfo); 
        if ((this.value = value) == null) {
            throw new NullPointerException("Value cannot be null");
        }
    }
    
    /**
      * 
      */
    @Override
    public ConstantExpression<E> copy() {
        return this;
    }
    
    /**
      * @return {@code constant}
      */
    @Override
    public final String name() {
        return "constant";
    }
    
    /**
      * 
      */
    @Override
    public E evaluate(Context context) {
        LogFactory.getLog(this).println("Evaluating: ", this.asSymbol(context.reset()));
        return this.value;
    }
    
    /**
      * Returns {@code x}, where {@code x} is the value of 
      * this constant, or if type information is included,
      * {@code Type{x}}, where {@code Type} is the simple
      * name of the type.   
      */
    @Override
    public String asSymbol(Context context) {
        String value = this.value.toString();
        if (this.includeTypeInfo) {
            value = this.type().getSimpleName() + "{" + value + "}"; 
        }
        return value;
    }
    
}
