/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.interpreter;

/**
  * An <i>initialisable expression</i> represents an expression
  * that require several steps to be constructed, and hence must 
  * be {@link #initialise() initialised} before actual use. <p>
  * 
  * After {@link #isInitialised() initialisation}, the expression
  * is considered immutable. Trying to alter it or initialise it
  * again is an error. 
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @param  <E> The type of value the evaluation of this expression 
  *             produces.
  */
public interface InitialisableExpression<E> extends Expression<E> {

    /**
      * Returns true if this expression has been {@link #initialise() 
      * initialised}, and hence ready to be evaluated, false if not. <p>
      * 
      * If this expression has not been initialised when evaluated,
      * the evaluation will throw an exception.
      *    
      * @return True if initialised, false if not.
      */
    public boolean isInitialised();

    /**
      * Initialises this expression before evaluation. <p>
      * 
      * Contained initialisable expressions are <b>not</b> initialised!
      *  
      * @return This expression; never null.
      * @throws ExpressionException If this expression has already
      *         been initialised.
      */
    public InitialisableExpression<E> initialise() throws ExpressionException;
    
    /**
      * Copies this initialisable expression. <p>
      * 
      * Unlike {@link #asSymbol(Context)}, copying 
      * <b>cannot</b> handle cyclic expression references!
      * 
      * @return A copy of this initialisable expression; never null.
      */
    public InitialisableExpression<E> copy(); 
    
}
