/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.iterator;

import static dk.rode.thesis.strategy.SequencePolicy.IDENTITY;

import java.util.Iterator;

import dk.rode.thesis.decorator.SequenceDecorator;
import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.ParticipantUsage;
import dk.rode.thesis.meta.log.Log;
import dk.rode.thesis.meta.log.LogFactory;
import dk.rode.thesis.meta.model.PrimeSequence;
import dk.rode.thesis.meta.model.Sequence;

/**
  * An <i>iterable sequence</i> is a <b>bounded</b> 
  * {@link Sequence sequence} that can be accessed as any 
  * other iterable object. <p>
  * 
  * When iterating, the iterable sequence will be advanced! 
  * To avoid modification of the sequence, simply 
  * {@link Sequence#copy() copy} it before use. <p>
  * 
  * <b>Example usage</b>:
  * <pre>
  *   // Highest number to consider as a prime number...
  *   int n = 100;
  *   // Create a bounded sequence to deliver primes...
  *   {@link PrimeSequence} sequence = new {@link PrimeSequence#PrimeSequence(int) PrimeSequence}(n);
  *   
  *   // Create iterable sequence based on "sequence"...
  *   {@link IterableSequence}&lt;Integer> is = new {@link #IterableSequence(Sequence) IterableSequence}&lt;Integer>(sequence);
  *   
  *   // Print all primes between 2 and n (implicit usage of {@link #iterator()})...
  *   for (int prime : is) {
  *     {@link LogFactory}.{@link LogFactory#getLog() getLog}().{@link Log#println(Object...) println}("Prime: ", prime);
  *   }
  * </pre>
  * 
  * An iterable sequence is not thread-safe. 
  *
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @see    SequenceIterator
  * @see    ProcessableSequence
  * 
  * @param  <E> The type of values delivered by this iterable sequence.
  */
@ParticipantUsage(value = "Aggregate", type = Iterable.class)
@Participant("ConcreteAggregate") 
public class IterableSequence<E> extends SequenceDecorator<E> implements Sequence<E>, Iterable<E> {
    
    /**
      * Constructor, which creates this iterable sequence to iterate
      * through the sequence values from {@code sequence}, starting
      * from its current value. <p>
      * 
      * {@code sequence} must be {@link Sequence#bounded() bounded} and 
      * is not altered by this constructor, but {@link Sequence#copy() 
      * copied} before returned by the {@link #iterator()} method. If
      * not bounded, {@code iterator()} will thrown an exception. 
      * 
      * @param  sequence The sequence to make iterable; cannot be null.
      * @throws NullPointerException If {@code sequence} is null.
      * @throws IllegalArgumentException If {@code sequence} is unbounded.
      */
    public IterableSequence(final Sequence<E> sequence) {
        super(sequence, IDENTITY);
        if (!sequence.bounded()) {
            throw new IllegalArgumentException("Sequence must be bounded");
        }
    }

    /**
      * 
      */
    public IterableSequence<E> copy() {
        return new IterableSequence<E>(this.getSequence(true));
    }
    
    /**
      * Returns an iterator to iterate over this sequence. <p>
      * 
      * The returned iterator <b>will</b> advance this sequence
      * because this sequence is iterable!
      * 
      * @return A new iterator; never null.
      */
    public Iterator<E> iterator() {
        return new SequenceIterator<E>(this);
    }

}
