/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.iterator;

import java.util.Iterator;

import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.ParticipantUsage;
import dk.rode.thesis.meta.model.Sequence;
import dk.rode.thesis.meta.model.Sequence.State;
import dk.rode.thesis.strategy.ObjectPolicy;
import dk.rode.thesis.strategy.SequencePolicy;
import dk.rode.thesis.strategy.Stringable;
import dk.rode.thesis.strategy.StringablePolicy;

/**
  * A <i>sequence iterator</i> is an <b>external iterator</b> 
  * that can iterate over <b>bounded</b> {@link Sequence sequences}. <p>
  * 
  * The {@link #next()} method returns the result of 
  * {@link Sequence#next()} for the sequence represented by
  * the iterator. The {@link #hasNext()} method will 
  * return false as long as the bound has not been reached. 
  * The functionality offered by a sequence iterator to
  * traverse the sequence values is roughly equivalent to
  * the following idiom: 
  * <pre>
  *   {@link Sequence} sequence = ..; 
  *   while (sequence.{@link Sequence#state() state()} != {@link dk.rode.thesis.meta.model.Sequence.State#RESTART}) {
  *     // do something meaningful with returned value...
  *     sequence.{@link Sequence#next() next()};
  *   }
  * </pre>
  * 
  * The state of the sequence represented by the iterator may
  * or may be affected if the sequence is accessed concurrently, 
  * depending on the means of creation. <p>
  * 
  * A sequence iterator is not thread-safe. 
  *  
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @see    #SequenceIterator(Sequence)
  * @see    #SequenceIterator(SequenceIterator)
  * @see    SequenceIterator#iterator(Sequence)
  * @see    IterableSequence
  * @see    ProcessableSequence    
  * 
  * @param  <E> The type of values delivered by this iterator.
  */
@ParticipantUsage(value = "Iterator", type = Iterator.class)
@Participant("ConcreteIterator")
public class SequenceIterator<E> implements Iterator<E>, Stringable<Sequence<E>> { 

    /**
      * The sequence to iterate over. <p>
      * 
      * Never null.
      */
    final Sequence<E> sequence;
            
    /**
      * Constructor. <p>
      * 
      * Modifications to {@code sequence} <b>will</b> affect this
      * iterator!
      * 
      * @param  sequence The sequence to iterate over; cannot be null.
      * @throws NullPointerException If {@code sequence} is null.
      */
    public SequenceIterator(final Sequence<E> sequence) {
        if ((this.sequence = sequence) == null) {
            throw new NullPointerException("Sequence cannot be null");
        } else if (!this.sequence.bounded()) {
            throw new IllegalArgumentException("Sequence must be bounded");
        }
    }
    
    /**
      * Copy constructor. <p>
      * 
      * Modifications to {@code iterator} will <b>not</b> affect this
      * iterator!
      * 
      * @param  iterator The iterator to copy; cannot be null.
      * @throws NullPointerException If {@code iterator} is null.
      */
    public SequenceIterator(SequenceIterator<E> iterator) {
        this.sequence = iterator.sequence.copy();
    }
    
    /**
      * Creates a new iterator based on {@code sequence}. <p>
      * 
      * Modification to {@code sequence} will <b>not</b> affect the
      * returned iterator as {@code sequence} is copied before use.
      * 
      * @param  <V>      The type of values delivered by {@code sequence}.
      * @param  sequence The sequence; cannot be null.
      * @return The created iterator; never null.
      * @throws NullPointerException If {@code sequence} is null.
      * 
      */
    public final static <V> Iterator<V> iterator(Sequence<V> sequence) {
        return new SequenceIterator<V>(sequence.copy());
    }
    
    /**
      * 
      */
    public boolean hasNext() {
        return this.sequence.state() != State.RESTART;
    }

    /**
      * 
      */
    public E next() {
        // Next actually corresponds to current...!
        E value = this.sequence.current();
        this.sequence.next();
        return value;
    }
    
    /**
      * @throws UnsupportedOperationException Always thrown.
      */
    public void remove() {
        throw new UnsupportedOperationException();
    }

    /**
      * 
      */
    public String toString() {
        return this.toString(null).toString();
    }

    /**
      * 
      */
    public StringablePolicy<? super Sequence<E>> getStringablePolicy(StringablePolicy<? super Sequence<E>> policy) {
        if (policy == null) {
            policy = SequencePolicy.SHORT;
        }
        return policy;
    }
    
    /**
      * 
      */
    public CharSequence toString(StringablePolicy<? super Sequence<E>> policy) {
        StringBuilder sb = (StringBuilder)ObjectPolicy.NAME_OPEN.toString(this);
        return sb.append(this.sequence.toString(this.getStringablePolicy(policy))).append(']').toString();
    }
    
}
