/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.memento;

import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.model.Sequence;
import dk.rode.thesis.meta.reflect.CallerClass;
import dk.rode.thesis.strategy.ObjectPolicy;
import dk.rode.thesis.strategy.SequencePolicy;
import dk.rode.thesis.strategy.Stringable;
import dk.rode.thesis.strategy.StringablePolicy;

/**
  * A <i>sequence memento</i> represents a <i>snap-shot</i> of a given
  * {@link Sequence sequence} at the time the {@link #SequenceMemento(Sequence)} 
  * constructor <b>or</b> the {@link #setSequence(Sequence)} method was
  * invoked. <p>
  * 
  * A memento can represent <b>any</b> type of sequence, but only 
  * {@link MemorizableSequence memorizable sequences} have a direct 
  * interface for getting and setting the sequence value based
  * on mementos. <p>
  *
  * Sequence mementos are thread-safe. <p>
  * 
  * <b>Implementation notes</b>:<br>
  * Sequence mementos uses the {@linkplain dk.rode.thesis.prototype Prototype} 
  * pattern to save the state of a given sequence by {@link Sequence#copy() 
  * copying} its state. <p>
  * 
  * All memento functionality has to be declared public if it must
  * be used outside this package. It cannot have "two interfaces" in Java, 
  * a narrow (public) and a wide (public and private), as described by Gamma 
  * et al. [<a href="http://www.rode.dk/thesis/Bibliography.htm#Gamma95" target="thesis-bibliography">Gamma95</a>, p.285-287]. This class corresponds to the private part of
  * the wide interface, while the {@link MemorizableSequence} corresponds to 
  * the narrow (public) interface. No access check is made in this class,
  * but the {@link GuardedSequenceMemento} provides an alternative, and tests 
  * if the {@link CallerClass caller} is a sequence of the proper type.
  *  
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @param  <E> The type of values delivered by applicable memorizable sequences.
  */
@Participant("Memento")
public class SequenceMemento<E> implements Stringable<Sequence<E>> {

    /**
      * A <b>local</b> (copy) sequence instance that stores 
      * internal state. <p>
      * 
      * Can be null until set. 
      */
    private Sequence<E> sequence;

    /**
      * No-arg constructor. <p>
      * 
      * The sequence representing the state to memorise must
      * be set explicitly using the {@link #setSequence(Sequence)}
      * method.
      */
    public SequenceMemento() {
        super();
    }
    
    /**
      * Constructor.
      * 
      * @param  sequence The sequence representing the state to 
      *                  memorise; cannot be null. This memento does
      *                  not keep a reference to {@code sequence},
      *                  but copies it.
      * @throws NullPointerException If {@code sequence} is null.
      */
    public SequenceMemento(final Sequence<E> sequence) {
        this();
        this.setSequence(sequence);
    }

    /**
      * Return the sequence memorised by this memento. <p>
      * 
      * Mutating the returned sequence will not alter the internal
      * state of this memento.
      *  
      * @return The memorised sequence; never null.
      * @throws IllegalStateException If no sequence has been set.
      * @see    #setSequence(Sequence)
      */
    public Sequence<E> getSequence() {
        if (this.sequence == null) {
            throw new IllegalStateException("No sequence has been set");
        }
        // Return copy because this memento is immutable...
        return this.sequence.copy();
    }

    /**
      * Returns the value stored in the memorised sequence.
      *  
      * @return The current (fixed) value; never null.
      * @throws IllegalStateException If no sequence has been set.
      */
    public E current() {
        if (this.sequence == null) {
            throw new IllegalStateException("No sequence has been set");
        }
        return this.sequence.current();
    }
    
    /**
      * Memorises the state of the {@code Sequence} supplied as 
      * sequence by copying it. The copy can then be retrieved 
      * using the {@link #getSequence()} method.
      * 
      * @param  sequence The sequence representing the state to 
      *                  memorise; cannot be null. This memento does
      *                  not keep a reference to {@code sequence},
      *                  but copies it.
      * @throws NullPointerException If {@code sequence} is null.
      * @throws IllegalStateException If a sequence has already been
      *         set.
      */
    public synchronized void setSequence(Sequence<E> sequence) {
        if (sequence == null) {
            throw new NullPointerException("Sequence cannot be null");
        } else if (this.sequence != null) {
            throw new IllegalStateException("Sequence already set");
        }
        // Defensive copy, so no other context can modify it by 
        // accessing "sequence"...
        this.sequence = sequence.copy();
    }
    
    /**
      * Returns the string representation of this memento.
      * 
      * @return The string representation; never null.
      */
    public String toString() {
        return this.toString(null).toString();
    }

    /**
      *  
      */
    public StringablePolicy<? super Sequence<E>> getStringablePolicy(StringablePolicy<? super Sequence<E>> policy) {
        if (policy == null) {
            return SequencePolicy.SHORT;
        }
        return policy;
    }

    /**
      * 
      */
    public CharSequence toString(StringablePolicy<? super Sequence<E>> policy) {
        StringBuilder sb = (StringBuilder)ObjectPolicy.NAME_OPEN.toString(this);
        return sb.append(this.sequence.toString(this.getStringablePolicy(policy)))
                 .append(']').toString();
    }
    
}
