/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.meta.log;

/**
  * The <i>log factory</i> is the public interface to create and 
  * access {@link Log logs}. <p> 
  *
  * <b>Implementation notes</b>:<br>
  * Internally, the factory uses a {@link LogManager} according to 
  * the {@linkplain dk.rode.thesis.singleton Singleton} pattern.
  * 
  * If the system property {@code dk.rode.thesis.log} is set to
  * {@code file}, file logs will be created by this manager. 
  * Otherwise the default system out loggers. <p>
  *  
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @see    LogManager
  */
public class LogFactory {

    /**
      * The singleton log manager that handles the logs. <p>
      * 
      * If the system property {@code dk.rode.thesis.log} is set to
      * {@code file}, file logs will be created by this manager. 
      * Otherwise the default system out loggers. <p>
      * 
      * Never null.
      */
    private static LogManager manager; 
    static {
        // Generally, I disapprove of using system properties to determine
        // behaviour, because it can have unpredictable results...
        String file = System.getProperty("dk.rode.thesis.log", "out").trim().toLowerCase();
        manager = new LogManager(file.equals("file") ? FileLog.class : AppendableLog.class, true);
    }

    /**
      * Private no-arg constructor.
      */
    private LogFactory() {
        super();
    }

    /**
      * Returns the global log.
      * 
      * @return The global log; never null.
      */
    public final static Log getLog() {
        return getLog(null);
    }

    /**
      * Returns the log associated with the class of the
      * object supplied supplied as {@code object}, or the 
      * global log if {@code object} is null. <p>
      * 
      * The class is tested on identity, and a given super or
      * sub-class will not match.
      * 
      * @param  object The object to deliver the class; can be null
      *                for the global log.
      * @return The log to use; never null.
      */
    public final static Log getLog(Object object) {
        return getLog(object == null ? null : object.getClass());
    }

    /**
      * Returns the log associated with the class supplied as 
      * {@code clazz}, or the global log if {@code clazz} is null. <p>
      * 
      * The class is tested on identity, and a given super or
      * sub-class will not match.
      * 
      * @param  clazz The class to which the log is associated; can be
      *               null for the global log.
      * @return The log to use; never null.
      */
    public final static Log getLog(Class<?> clazz) {
        return manager.getLog(clazz);
    }
        
}
