/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 * 
 */
package dk.rode.thesis.meta.model;

import dk.rode.thesis.strategy.SequencePolicy;
import dk.rode.thesis.strategy.StringablePolicy;

/**
  * An <i>abstract sequence</i> implements the basic traits of 
  * the {@link Sequence} interface. <p>
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @param  <E> The type of values delivered by this sequence. 
  */
public abstract class AbstractSequence<E> implements Sequence<E> {

    /**
      * The internal state. <p>
      * 
      * Never null.
      */
    protected State state;
    
    /**
      * No-arg constructor, which {@link #reset() reset} this sequence
      * and then initialise the internal state to {@link Sequence.State#START}. <p>
      * 
      * Note, that this sequence may be reset before completely
      * initialised!
      */
    protected AbstractSequence() {
        this(true);
    }

    /**
      * Constructor, which {@link #reset() reset} this sequence
      * if {@code reset} is true and then initialises the internal
      * state to {@link Sequence.State#START}. <p>
      * 
      * Note, that this sequence can be reset before completely
      * initialised!
      * 
      * @param reset True to reset.
      */
    protected AbstractSequence(boolean reset) { 
        if (reset) {
            this.reset();
        }
        this.state = State.START;
    }
    
    /**
      * Copy constructor. <p>
      * 
      * The state of {@code sequence} is set for this sequence.
      * 
      * @param  sequence The sequence to copy; cannot be null.
      * @throws NullPointerException If {@code sequence} is null.
      */
    protected AbstractSequence(Sequence<? super E> sequence) {
        this.state = sequence.state();
    }
    
    /**
      * 
      */
    public void reset() {
        this.state = State.RESET;
    }
    
    /**
      * 
      */
    public State state() {
        return this.state;
    }
    
    /**
      * Returns the default string representation specified by
      * the {@link #toString(StringablePolicy)} method.
      *
      * @return The string representation; never null.
      */
    public String toString() {
        // By using null here, sub-class need only override
        // toString(..) to change the default behaviour...
        return this.toString(null).toString();
    }

    /**
      * 
      */
    public StringablePolicy<? super Sequence<E>> getStringablePolicy(StringablePolicy<? super Sequence<E>> policy) {
        if (policy == null) {
            policy = SequencePolicy.VERBOSE;
        }
        return policy;
    }

    /**
      * 
      */
    public CharSequence toString(StringablePolicy<? super Sequence<E>> policy) {
        return this.getStringablePolicy(policy).toString(this);
    }

}
