/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.meta.reflect.proxy;

import java.lang.reflect.Method;

/**
  * A <i>reference</i> is used to handle the number of references
  * to a given shared object: as long as only <i>non-mutable</i> 
  * methods are invoked on the referenced object, the object can be 
  * shared among several proxies. Once a given <i>mutator</i> method 
  * is invoked, the referenced object will be copied using a 
  * specified <i>copy</i> method. <p>
  * 
  * <b>Implementation notes</b>:<br>
  * This corresponds to a <i>smart reference</i> (or <i>smart pointer</i>)
  * as described by Gamma et al. [<a href="http://www.rode.dk/thesis/Bibliography.htm#Gamma95" target="thesis-bibliography">Gamma95</a>, p.209] and is thus an application
  * of the {@linkplain dk.rode.thesis.proxy Proxy} pattern. It works via
  * {@link java.lang.reflect.Proxy java.lang.reflect.Proxy} objects. It 
  * corresponds to the Handle/Body C++ idiom described by Coplien.
  *
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @see    ReferenceInvocationHandler
  * @see    ProxyFactory#getSharedObject(Object, String, Class[], String...)
  * 
  * @param  <T> The type of the referenced object.
  */
public interface Reference<T> {

    /**
      * Returns the referenced object. 
      *  
      * @return The referenced object; never null.
      */
    public T getReference();

    /**
      * Increases the reference count for the 
      * {@link #getReference() referenced} object.
      *   
      * @return The referenced object; never null.
      */
    public T addReference();
    
    /**
      * Returns true if the method supplied as {@code method}
      * is the <i>copy</i> method used to copy the referenced 
      * object, false if not.
      *  
      * @param  method    The method to test; cannot be null.
      * @return True if {@code method} is the method used
      *         to copy objects of the referenced object type,
      *         false if not. 
      * @throws NullPointerException If {@code method} is null.         
      */
    public boolean copyMethod(Method method);
    
    /**
      * Performs the actual copying of the referenced object if
      * {@code method} is a <i>mutator</i> method and returns
      * a new reference containing the copy with a reference
      * count of one. The reference count for this reference
      * will at the same time be decreased by one. <p>
      * 
      * If {@code method} is not a <i>mutator</i> method, this
      * reference is simply returned.
      * 
      * @param  method The method to test; cannot be null.
      * @return The reference to use; never null.
      * @throws NullPointerException If {@code method} is null.
      * @throws Exception If the copying fails.
      */
    public Reference<T> copyIfMutator(Method method) throws Exception;

}
