/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.meta.util;

import static dk.rode.thesis.meta.util.Throwables.toRuntimeException;

import java.io.BufferedInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.util.Arrays;
import java.util.Collections;
import java.util.IdentityHashMap;
import java.util.List;
import java.util.Set;
import java.util.concurrent.atomic.AtomicInteger;

/**
  * Various <i>utility functions</i>.
  *
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  */
public class Util {

    /**
      * Private off-limit constructor.
      */
    private Util() {
        throw new AssertionError(); // guard against reflection
    }

    /**
      * Creates a new array with {@code value} at index 0,
      * and {@code values} from index 1 to {@code values.length},
      * in order. <p>
      * 
      * Null values are handled like any other value.
      *   
      * @param  value  The first value in the new array; can be null.
      * @param  values The last values, in order, in the new array; can 
      *                contain null entries.
      * @return A new array containing all the values; never null.
      */
    public final static Object[] shift(Object value, Object... values) {
        Object[] array = new Object[values.length + 1];
        array[0] = value;
        System.arraycopy(values, 0, array, 1, values.length);
        return array;
    }
    
    /**
      * Adds each element in {@code elements}, in order, to the list
      * supplied as {@code list}. <p>
      * 
      * Null elements in {@code elements} will be ignored.
      * 
      * @param  <E>      The type of elements.
      * @param  list     The list; cannot be null.
      * @param  elements The elements to add to {@code list}; can contain
      *                  null entries, which will be ignored.
      * @return {@code list}; never null.
      * @throws NullPointerException If {@code list} is null.
      */
    public final static <E> List<E> add(List<E> list, E... elements) {
        if (elements == null) {
            return list;
        }
        for (E element : elements) {
            if (element != null) {
                list.add(element);
            }
        }
        return list;
    }

    /**
      * Adds each element in {@code elements}, in order, to the set
      * supplied as {@code set}. <p>
      * 
      * Null elements in {@code elements} will be ignored.
      * 
      * @param  <E>      The type of elements.
      * @param  set      The set; cannot be null.
      * @param  elements The elements to add to {@code set}; can be null or 
      *                  can contain null entries, which will be ignored.
      * @return {@code set}; never null.
      * @throws NullPointerException If {@code set} is null.
      */
    public final static <E> Set<E> add(Set<E> set, E... elements) {
        if (elements == null) {
            return set;
        }
        for (E element : elements) {
            if (element != null) {
                set.add(element);
            }
        }
        return set;
    }
    
    /**
      * Copies all elements from {@code from} into {@code to},
      * from index zero.
      *  
      * @param  <E>  The type of elements stored in {@code to}.
      * @param  from The from array; cannot be null.
      * @param  to   The to array; cannot be null.
      * @return The {@code to} array; never null.
      * @throws ArrayStoreException If elements in {@code from} cannot
      *         be stored in {@code to}.
      */
    public final static <E> E[] copy(Object[] from, E[] to) {
        return copy(from, to, from.length);
    }

    /**
      * Copies the first {@code length} elements from {@code from} 
      * into {@code to}, from index zero.
      *  
      * @param  <E>    The type of elements stored in {@code to}.
      * @param  from   The from array; cannot be null.
      * @param  to     The to array; cannot be null.
      * @param  length The number of elements to copy.
      * @return The {@code to} array; never null.
      * @throws IndexOutOfBoundsException If {@code length} is illegal.         
      * @throws ArrayStoreException If elements in {@code from} cannot
      *         be stored in {@code to}.
      */
    public final static <E> E[] copy(Object[] from, E[] to, int length) { 
        System.arraycopy(from, 0, to, 0, length);
        return to;
    }
    
    /**
      * Returns true if {@code a} equals {@code b}, including
      * true if both {@code a} and {@code b} are null. <p>
      * 
      * Array types are compared based on a proper {@code Arrays.equals(..)}
      * method, including support for {@link Primitive primitives}.
      *   
      * @param  a The first object; can be null.
      * @param  b The second object; can be null.
      * @return True if equal, false if not.
      */
    public final static boolean equals(Object a, Object b) {
        if (a == null) {
            return (b == null);
        } else if (b == null) {
            return false;
        } 
        Class<?> aClass = a.getClass(), bClass = b.getClass();
        if ((aClass.isArray()) && (bClass.isArray())) {
            aClass = aClass.getComponentType(); bClass = bClass.getComponentType();
            boolean aPrimitive = aClass.isPrimitive(), bPrimitive = bClass.isPrimitive();  
            if ((aPrimitive) && (bPrimitive)) {
                if (aClass.equals(bClass)) {
                    switch (Primitive.getPrimitive(aClass)) {
                        case Boolean:   return Arrays.equals((boolean[])a, (boolean[])b);
                        case Byte:      return Arrays.equals((byte[])a,    (byte[])b);
                        case Character: return Arrays.equals((char[])a,    (char[])b);
                        case Double:    return Arrays.equals((double[])a,  (double[])b);
                        case Float:     return Arrays.equals((float[])a,   (float[])b);
                        case Integer:   return Arrays.equals((int[])a,     (int[])b);
                        case Long:      return Arrays.equals((long[])a,    (long[])b);
                        case Short:     return Arrays.equals((short[])a,   (short[])b);
                    }
                }
                // Fall-though
                
            // Cannot cast primitives into Object[]...
            } else if ((!aPrimitive) && (!bPrimitive)) {
                return Arrays.equals((Object[])a, (Object[])b);
            }
        }
        return a.equals(b);
    }
    
    /**
      * Returns true {@code first} equals all other objects in
      * {@code rest}, false if not. <p>
      * 
      * The comparison between two objects is performed using
      * the {@link #equals(Object, Object)} method. <p>
      * 
      * If {@code rest} is empty, true is returned.
      * 
      * @param  first The first object; can be null.
      * @param  rest  The rest of the objects; can be null or empty.
      * @return True if all objects are equal, false if not.
      */
    public final static boolean equals(Object first, Object... rest) {
        for (Object o : rest) {
            if (!equals(first, o)) {
                return false;
            }
        }
        return true;
    }
    
    /**
      * To make temporary file names unique during serialization copy. 
      */
    private final static AtomicInteger FILE_ID = new AtomicInteger(0);
    
    /**
      * Copies {@code object} by serializing it and returning the
      * deserialized copy. <p>
      * 
      * <b>Implementation notes</b>:<br>
      * This implementation is <b>not</b> fast, as it utilises a temporary
      * file to store the serialized data. I truly doubt serialization was
      * invented for this purpose... :)
      * 
      * @param  <T>    The type of serializable object.
      * @param  object The object to copy; cannot be null.
      * @return A copy of {@code object}; never null.
      * @throws NullPointerException If {@code object} is null.
      * @throws RuntimeException If the copying (serialization) fails.
      */
    public final static <T extends Serializable> T serializableCopy(T object) {
        try {
            File file = File.createTempFile("serializeableCopy" + FILE_ID.getAndIncrement() + ".ser", null);

            ObjectOutputStream oos = new ObjectOutputStream(new FileOutputStream(file));
            oos.writeObject(object);
            oos.close();

            ObjectInputStream ios = new ObjectInputStream(new FileInputStream(file));
            // Re-read object known to be of type T...
            @SuppressWarnings("unchecked") T copy = (T)ios.readObject();
            ios.close();
            
            return copy;
            
        } catch (Exception e) {
            throw toRuntimeException(e);
        }
    }
    
    /**
      * Returns the data directory used for test purposes.
      *   
      * The directory is named {@code data} and is placed 
      * in the root directory of the class path.
      * 
      * @return The directory; never null.
      */
    public final static File getDataDirectory() {
        return getDirectory("data", false); 
    }

    /**
      * Returns the log directory used for test purposes. <p>
      * 
      * The directory is named {@code log} and is placed 
      * in the root directory of the class path. The directory
      * is created if it does not exist.
      *   
      * @return The directory; never null.
      */
    public final static File getLogDirectory() {
        return getDirectory("log", true);
    }

    /**
      * Returns the directory named {@code name} from the root
      * of the class path. <p>
      * 
      * If {@code create} is true and no such directory exist, it
      * will be created.
      * 
      * @param  name   The directory name; cannot be null.
      * @param  create True to create if no such directory exist.
      * @return The directory; never null.
      * @throws NullPointerException If {@code name} is null.
      * @throws IllegalArgumentException If {@code name} does not denote a 
      *         (created) directory.
      */
    public final static File getDirectory(String name, boolean create) {
        if (name == null) {
            throw new NullPointerException("Directory name cannot be null");
        }
        // ISSUE: if the Eclipse project *links* to the source files, the 
        //        location will not be as expected and will cause an error
        //        when the "data" directory is tried accessed...!
        File file = new File((File)null, name).getAbsoluteFile();
        if (create) {
            if (!file.exists()) {
                file.mkdir();
            }
        }
        if (!file.isDirectory()) {
            throw new IllegalArgumentException("Name does not denote a directory: " + name);
        }
        return file;
    }
    
    /**
      * Reads the contents of the input stream supplied as {@code in}
      * and returns the contents as a byte array. <p>
      * 
      * {@code in} is not closed by this method.
      * 
      * @param  in The input stream; cannot be null.
      * @return A byte array representing the contents of {@code in}; 
      *         never null.
      * @throws NullPointerException If {@code in} is null.        
      * @throws IOException If the reading fails.
      */
    public final static byte[] readBytes(InputStream in) throws IOException {
        BufferedInputStream bis = null;
        if (in instanceof BufferedInputStream) {
            bis = (BufferedInputStream)in;
        } else {
            bis = new BufferedInputStream(in);
        }
        ByteArrayOutputStream baos = new ByteArrayOutputStream();
        // Internal buffer...
        final byte[] buffer = new byte[512];
        int read = 0;
        while ((read = bis.read(buffer)) > 0) {
            baos.write(buffer, 0, read);
        }
        // No need to close a byte array output stream...
        return baos.toByteArray();
    }
    
    /**
      * Returns a new <b>identity</b> hash set.
      *  
      * @param  <E> The type of elements stored in the set.
      * @return A new identity hash set; never null.
      */
    public final static <E> Set<E> newIdentityHashSet() {
        return Collections.newSetFromMap(new IdentityHashMap<E,Boolean>()); // ISSUE: Does not compile using Java 5       
    }
    
}
