/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.observer;

import java.util.Collection;

import dk.rode.thesis.adapter.SequenceAdapter;
import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.log.Log;
import dk.rode.thesis.meta.log.LogFactory;
import dk.rode.thesis.meta.model.AbstractSequence;
import dk.rode.thesis.meta.model.Sequence;
import dk.rode.thesis.meta.model.Sequence.State;
import dk.rode.thesis.meta.reflect.Executor;
import dk.rode.thesis.observer.ObserverManager.NotificationPolicy;
import dk.rode.thesis.proxy.ImmutableSequence;
import dk.rode.thesis.strategy.SequencePolicy;

/**
  * An <i>annotated observers sequence</i> implements the basic
  * traits of any {@link ObservableSequence observable} sequence 
  * that uses the {@link Executor} annotation to identify notification 
  * methods for observers that accepts a {@code Sequence} type as the
  * first argument and a {@link dk.rode.thesis.meta.model.Sequence.State}
  * type as the second. <p>
  * 
  * The context {@link Executor#value() type} must furthermore be a type 
  * this sequence type is assignable to (default), or an explicit type
  * supplied at {@link #AnnotatedObserversSequence(Class) construction}
  * time. <p>
  * 
  * The {@link ObservableSequence observable} sequence 
  * functionality is fully implemented by this class! Sub-classes
  * need simply implement the {@link #doNext()} method
  * (as opposed to {@link #next()}) without worrying about
  * notifying observers. <p>
  * 
  * This sequence is <b>not</b> thread-safe! It does not enforce
  * synchronisation, neither when adding or removing observers, nor
  * when observers are notified. It could be potentially dangerous
  * or least time consuming to notify observers from a synchronised 
  * context, holding a lock on the sequence itself. If synchronisation
  * is required, the context using the sequence must enforce it. <p>
  * 
  * <b>Implementation notes</b>:<br>
  * This sequence uses an {@link ObserverManager} internally to 
  * handle the observer functionality. <p>
  *
  * An observer can be notified any method accepting a {@code Sequence} 
  * as the first argument and a {@code Sequence.State} type as the second. 
  * This corresponds to the {@link SequenceObserver#sequenceEvent(Sequence, Object)
  * notification method} defined in the {@link SequenceObserver} interface, 
  * but could easily have been different by supplying different formal
  * parameter types to the {@code ObserverManager} used.
  * 
  * This class applies the {@linkplain dk.rode.thesis.templatemethod Template Method}
  * pattern: {@link #next()} is the template method, and {@code doNext()} is a 
  * primitive operation. This ensures consistent state during notification.
  *  
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @see    ObserverManager
  * @see    Executor
  * 
  * @param  <E> The type of values delivered by this sequence. 
  */
@Participant("Subject")
public abstract class AnnotatedObserversSequence<E> 
              extends AbstractSequence<E> 
           implements ObservableSequence<Object,State,E> {

    /**
      * The observer manager storing the observers. <p>
      * 
      * Never null.
      */
    private final ObserverManager observers;

    /**
      * No-arg sub-class constructor. <p>
      * 
      * The context {@link Executor#value() type} of the 
      * {@link Executor} annotations must be assignable to 
      * the type of this sequence in order to be accepted
      * as notification methods by this observable sequence.
      *                          
      * @see    ObserverManager#getOwnerType()                       
      */
    protected AnnotatedObserversSequence() {
        this((Class<?>)null);
    }
    
    /**
      * Sub-class constructor. <p>
      * 
      * The context {@link Executor#value() type} of the 
      * {@link Executor} annotations must be assignable to 
      * the type supplied as {@code ownerType} in order to be accepted
      * as notification methods by this observable sequence.
      *                          
      * @param  ownerType The type to use; can be null, in which case the
      *                   type of this sequence is used.
      * @see    ObserverManager#getOwnerType()                       
      */
    protected AnnotatedObserversSequence(Class<?> ownerType) {
        super(false); 
        this.observers = new ObserverManager(NotificationPolicy.ContinueOnError,
                                             ownerType == null ? this.getClass() : ownerType, 
                                             Sequence.class, State.class);
    }

    /**
      * Copy constructor.
      * 
      * @param  sequence The observable sequence to copy; cannot be null.
      * @throws NullPointerException If {@code sequence} is null.
      */
    public AnnotatedObserversSequence(AnnotatedObserversSequence<E> sequence) {
        super(sequence);
        // Copy observer references...
        this.observers = new ObserverManager(sequence.observers);
    }
    
    /**
      * {@inheritDoc}
      * 
      * Observers associated with this observable sequence are notified
      * with the (new) state of this sequence.
      * 
      * @see #sequenceEvent(Sequence, Sequence.State)
      */
    public final E next() { // Template method
        // Fetch value via sub-class hook, which may alter the state...
        E value = this.doNext();
        // Notify, always...
        this.sequenceEvent(this, this.state);
        return value;
    }

    /**
      * Hook for sub-classes to perform the actual {@code next()}
      * operation <b>without</b> notifying observers.
      *  
      * @return The next element; never null.
      */
    protected abstract E doNext(); // Primitive operation
    
    /**
      * 
      */
    public Collection<Object> getObservers() {
        return this.observers.getObservers();
    }
    
    /**
      * 
      */
    public boolean addObserver(@Participant("ConcreteObserver") Object object) {
        return this.observers.addObserver(object);
    }

    /**
      * 
      */
    public boolean removeObserver(Object object) {
        return this.observers.removeObserver(object);
    }

    /**
      * 
      */
    public void sequenceEvent(Sequence<?> sequence, State state) {
        // When we notify observers, we wrap this sequence in an 
        // immutable proxy sequence so it cannot be altered by 
        // observers. This will also ensure that notifications are 
        // issued only by the actual client...
        Log out = LogFactory.getLog(this);
        
        // Notify...
        if (sequence == this) {
            if (this.observers.size() > 0) {
                // To create an immutable sequence, we have to adapt the actual type... 
                ImmutableSequence<?> is = 
                    new ImmutableSequence<Object>(SequenceAdapter.create(this));
                // Supply this sequence (decorated) and the state to each notified 
                // observer...
                this.observers.notifyObservers(is, state);
            }
            
        // Notified...
        } else {
            out.println("-> ", this.toString(SequencePolicy.ID), " notified about: ", sequence.toString(SequencePolicy.ID), ": ", state);
        }
    }
    
}
