/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.observer;

import java.util.Calendar;
import java.util.Date;

import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.model.Sequence;
import dk.rode.thesis.meta.model.Sequence.State;

/**
  * An <i>observable date sequence</i> is a sequence that will 
  * advance an initial date by one day each time {@link #next() next()} is 
  * invoked and then notify relevant {@link SequenceObserver observers}
  * of its state changes. <p>
  *  
  * The returned dates only contain hours, day, month, and year. <p>
  * 
  * Modifying returned dates will not affect a date sequence. <p>
  * 
  * A date sequence may be {@link Sequence#bounded() bounded}, but is
  * {@link Sequence#consistent() consistent} and {@link #unique() unique}.
  *
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  */
@Participant({"ConcreteSubject", "ConcreteObserver"})
public class DateSequence 
     extends SequenceObserversSequence<Date,State> {

    /**
      * The start date of this sequence. <p> 
      *  
      * Invariant: start <= {@link #date} [<= {@link #end}]. <p>
      * 
      * Never null.
      */
    private final Date start;
    /**
      * The end date of this sequence, if any. Null means no
      * end date. <p> 
      *  
      * Invariant: {@link #start} <= {@link #date} [<= end] 
      */
    private final Date end;
    /**
      * The current date in milliseconds. <p>
      * 
      * Hours, minutes, seconds, and milliseconds are ignored,
      * i.e. always zero. <p>
      * 
      * Never null.
      */
    private long date;
    
    /**
      * No-arg constructor. <p>
      * 
      * The start date is set to "now", and no end date is used.
      */
    public DateSequence() {
        this(new Date());
    }

    /**
      * Constructor. <p>
      * 
      * This date sequence will not have an end date.
      * 
      * @param  start The date to start this sequence from; cannot be null.
      *               Only hours, days, months, and years are used.
      * @throws NullPointerException If {@code start} is null.              
      */
    public DateSequence(Date start) {
        this(start, null);
    }

    /**
      * Constructor. <p>
      * 
      * This date sequence will have an end date {@code days}
      * from {@code start}. If {@code days} is smaller than one, one
      * day will be used.
      * 
      * @param  start The date to start this sequence from; cannot be null.
      *               Only hours, days, months, and years are used.
      * @param  days  The number of days in this sequence.               
      * @throws NullPointerException If {@code start} is null.
      */
    public DateSequence(Date start, int days) {
        this(start, getDate(start, days < 1 ? 1 : days));
    }
    
    /**
      * Constructor. <p>
      *
      * If {@code end} is null, this date sequence is unbounded.
      * 
      * @param  start The date to start this sequence from; cannot be null.
      *               Only hours, days, months, and years are used.
      * @param  end   The end date of this sequence; can be null. Only hours,
      *               days, months, and years are used.              
      * @throws NullPointerException If {@code start} is null.
      * @throws IllegalArgumentException If {@code end} is before 
      *         {@code start}.              
      */
    public DateSequence(Date start, Date end) {
        super(); 
        if (start == null) {
            throw new NullPointerException("Start date cannot be null");
        } else if ((end != null) && (end.before(start))) {
            throw new IllegalArgumentException("End date cannot be before start date");
        }
        // Defensive copying...
        this.start = getDate(start);
        this.end   = end == null ? null : getDate(end);
        this.date  = this.start.getTime();
    }

    /**
      * Copy constructor.
      * 
      * @param  sequence The date sequence to copy; cannot be null.
      * @throws NullPointerException If {@code sequence} is null.
      */
    public DateSequence(DateSequence sequence) {
        super(sequence);
        this.start = sequence.start;
        this.end   = sequence.end;
        this.date  = sequence.date;
    }

    /**
      * 
      */
    public DateSequence copy() {
        return new DateSequence(this);
    }

    /**
      *  
      */
    public State getAspect() {
        return this.state; // want update every time next or reset is invoked...
    }
    
    /**
      * 
      */
    public boolean bounded() {
        return (this.end != null);
    }

    /**
      * Returns true.
      * 
      * @return True.
      */
    public boolean consistent() {
        return true;
    }

    /**
      * Returns true.
      * 
      * @return True.
      */
    public boolean unique() {
        return true;
    }
    
    /**
      * 
      */
    public void reset() {
        super.reset();
        this.sequenceEvent(this, this.state);
    }
    
    /**
      * 
      */
    public Date current() {
        return new Date(this.date);
    }

    /**
      * 
      */
    protected Date doNext() {
        Calendar calendar = Calendar.getInstance();
        calendar.setTimeInMillis(this.date);
        calendar.add(Calendar.DATE, 1);
        Date date = calendar.getTime();
        // Bounded?
        if ((this.end != null) && (date.after(this.end))) {
            date = this.start;
            this.state = State.RESTART;
        } else {
            this.state = State.NORMAL;
        }
        this.date = date.getTime();
        return date;
    }

    /**
      * Returns a date corresponding to {@code date}, but
      * where minutes, seconds, and milliseconds
      * have been cleared.
      *  
      * @param  date The date; cannot be null.
      * @return A new corresponding {@code Date} instance; 
      *         never null.
      */
    private final static Date getDate(Date date) {
        return getDate(date, 0);
    }

    /**
      * Returns a date corresponding to {@code date} plus
      * the number of days supplied as {@code days}, but
      * where minutes, seconds, and milliseconds
      * have been cleared.
      *  
      * @param  date The date; cannot be null.
      * @param  days Days to add to the date; cannot be negative.
      * @return A new corresponding {@code Date} instance; 
      *         never null.
      * @throws IllegalArgumentException If {@code days} is negative.         
      */
    private final static Date getDate(Date date, int days) {
        if (days < 0) {
            throw new IllegalArgumentException("Days cannot be negative: " + days);
        }
        Calendar calendar = Calendar.getInstance();
        calendar.setTime(date);
        calendar.clear(Calendar.MINUTE);
        calendar.clear(Calendar.SECOND);
        calendar.clear(Calendar.MILLISECOND);
        if (days > 0) {
            calendar.add(Calendar.DATE, days);
        }
        return calendar.getTime();
    }
        
}
