/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.prototype;

import dk.rode.thesis.interpreter.Expression;
import dk.rode.thesis.interpreter.TypedExpression;
import dk.rode.thesis.meta.Participant;

/**
  * A <i>copyable</i> object can return a <i>deep copy</i> of it self  
  * any type implementing this interface can act as a <i>prototype</i>. <p>
  * 
  * Though not required, it is prudent to supply a <i>copy constructor</i>
  * for any type implementing the {@code Copyable} interface. <p>
  *
  * <b>Implementation notes</b>:<br>
  * The {@code Copyable} interface is generic, accepting the type parameter
  * {@code T} that determines the type of the returned copy. The type parameter 
  * is used for type safety reasons (as opposed to simply returning {@code Object}). 
  * Though it cannot be enforced, {@code T} is expected to be the implementing 
  * type itself. If a bound like {@code <T extends Copyable<T>>} had been 
  * used instead of just {@code T}, the {@link PrototypeFactory} would not have
  * been able to generate <i>dynamic proxies</i> to represent prototypical
  * objects that dynamically implements {@code Copyable} for <i>any</i> object 
  * type: the original type(s) could not be guaranteed to satisfy the 
  * {@code extends Copyable} bound. If such a bound is to be enforced,
  * the {@link StrictCopyable} type can be used. <p>
  * 
  * Sub-interfaces can use <i>covariant return types</i> to refine the type
  * of the returned copy even further. This is sometimes even required. An
  * example of this is the {@link Expression} and {@link TypedExpression} 
  * interfaces used in the {@linkplain dk.rode.thesis.interpreter Interpreter} 
  * pattern. {@code TypedExpression} extends {@code Expression} to augment
  * normal expressions with runtime type information. If it had not overridden
  * the {@code copy()} method from {@code Expression} that returns an
  * {@code Expression}, copying a {@code TypedExpression} would return
  * an {@code Expression}, not a {@code TypedExpression}. Hence, the 
  * compile-time behaviour of {@code TypedExpression} is not accessible
  * without explicit casting if copied. By declaring the {@code copy()}
  * method to return a {@code TypedExpression} no casting is required. 
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  *
  * @see    PrototypeFactory
  * 
  * @param  <T> The type of copyable object. 
  */
@Participant("Prototype")
public interface Copyable<T> {

    /**
      * Returns a <i>deep copy</i> of this object.
      *  
      * @return The copy; never null.
      */
    public T copy();
    
}
