/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 * 
 */
package dk.rode.thesis.prototype;

import static dk.rode.thesis.decorator.DecoratedPolicy.VISIBLE;

import java.lang.reflect.InvocationHandler;

import dk.rode.thesis.decorator.AppenderDecorator;
import dk.rode.thesis.iterator.SequenceIterator;
import dk.rode.thesis.memento.RangeSequence;
import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.log.Log;
import dk.rode.thesis.meta.log.LogFactory;
import dk.rode.thesis.meta.model.EnglishAlphabetSequence;
import dk.rode.thesis.meta.model.LongSequence;
import dk.rode.thesis.meta.model.Sequence;
import dk.rode.thesis.meta.reflect.TypeLiteral;
import dk.rode.thesis.meta.test.Testable;
import dk.rode.thesis.meta.util.Arguments;
import dk.rode.thesis.strategy.ObjectPolicy;

/**
  * <b>Prototype</b> tests.
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  */
@Participant("Client")
public class Main implements Testable {

    /**
      * Executes the <b>Prototype</b> tests.
      * 
      * @param args The arguments to supply to {@link #test(Log, Arguments)},
      *             if any; can be null.
      */
    public static void main(String[] args) {
        new Main().test(LogFactory.getLog(), new Arguments(args));
    }

    /**
      * No-arg constructor.
      */
    public Main() {
        super();
    }
    
    /**
      * 
      */
    public boolean test(Log out, Arguments arguments) {
        out.heading("Prototype tests...");
       
        if (arguments.size() > 0) {
            out.println("Supplied arguments: ", arguments);
        }
        
        // Invariant: the "copy()" functionality defined in
        //            the Sequence interface makes any sequence
        //            a prototype.

        PrototypeFactory pf = new PrototypeFactory();
        
        // True for verbose logging...
        boolean log = arguments.getArgumentValue("log.verbose", false, Boolean.class);
        LogFactory.getLog(InvocationHandler.class).setActive(log);
        
        // Sequences that acts as prototypes for new sequences...
        Sequence<?>[] sequences = {
            new LongSequence(),
            // All sequences already implement "copy()", so the sequence
            // will just be used as is...
            pf.getPrototypicalObject(SymbolSequence.class, new SymbolSequence()),
            new SymbolSequence('(', '[', '<', '>', ']', ')'),
            new RangeSequence(42, 142),
            CountdownSequence.getInstance(),
            new AppenderDecorator(new EnglishAlphabetSequence(), "<", ">")
        };
                
        // Print instances...
        out.heading("Sequence instances");
        for (int i = 0; i < sequences.length; i++) {
            Sequence<?> sequence = sequences[i];
            out.println("Sequence ", i, ": ", sequence, " -> ", sequence.toString(VISIBLE));
        }
        
        out.heading("Sequence prototype usage");
        for (int i = 0; i < sequences.length; i++) {
            Sequence<?> prototype = sequences[i];
            // Use "sequence" as prototype to supply new sequences...
            Sequence<?> copy = prototype.copy();
            Sequence<?> clone = null;
            
            out.println();
            out.println("Sequence ", i, ":");
            out.println("Sequence ", i, " copied: ", prototype, " -> ", copy, " -> equal: ", prototype.equals(copy));
            
            // Built in clone mechanism also acts as a prototype. Notice
            // that java.lang.Cloneable does not define any public "clone()"
            // method, so the exact type must be known in order to invoke
            // it (unless reflection is used). We cast to an object before
            // we use instanceof because we use wild-cards in Sequence<?>...
            
            Object o = prototype;
            if (o instanceof SymbolSequence) {
                SymbolSequence ss = (SymbolSequence)o;
                try {
                    clone = ss.clone();
                    out.println("Sequence ", i, " cloned: ", prototype, " -> ", clone, " -> equal: ", prototype.equals(clone));
                    
                } catch (Throwable t) {
                    out.println("Sequence ", i, " cannot be cloned: ", prototype);
                    out.println(t);
                }
                
            } else if (o instanceof CountdownSequence) {
                CountdownSequence cs = (CountdownSequence)o;
                try {
                    clone = cs.clone();
                    out.println("Sequence ", i, " cloned: ", prototype, " -> ", clone, " -> equal: ", prototype.equals(clone));
                    
                } catch (Throwable t) {
                    out.println("Sequence ", i, " is cloneable, but cannot be cloned: ", prototype);
                    out.println(t);
                }
            }
            
            // Iterate the different sequences...
            out.println();
            out.println("-> Advancing prototype (2), copy (6), and possible clone (12)...");
            for (int j = 0; j < 2; j++) {
                prototype.next();
                for (int k = 0; k < 3; k++) {
                    copy.next();
                    if (clone != null) {
                        for (int l = 0; l < 2; l++) {
                            clone.next();
                        }
                    }
                }
            }
            out.println("   Prototype: ", prototype);
            out.println("   Copy:      ", copy);
            if (clone != null) {
                out.println("   Clone:     ", clone);
            }
        }

        out.println().heading("Sequence iterator instances (dynamic prototypes based on proxies)");
        
        // Not copyable, but defines a copy constructor...
        SequenceIterator<Integer> si = 
            new SequenceIterator<Integer>(new RangeSequence(1,9)); 

        out.println("Main sequence iterator: ", si, " = ", ObjectPolicy.ID.toString(si)).println();
        
        Copyable<?> proto1 = 
            (Copyable<?>)pf.getPrototypicalObject(SequenceIterator.class, si);

        // Null if no copy constructor...
        if (proto1 != null) {
            out.println("Sequence iterator 0 (class literal):   ", proto1, " = ", ObjectPolicy.ID.toString(proto1));
            Object si1 = proto1.copy();
            out.println("Sequence iterator 0 copy (anonymous): ", si1, " = ", ObjectPolicy.ID.toString(si1));
        }
        
        TypeLiteral<SequenceIterator<Integer>> literal = 
            new TypeLiteral<SequenceIterator<Integer>>(){/*magic*/};
        
        Copyable<SequenceIterator<Integer>> proto2 = 
            pf.getPrototypicalObject(literal, si);

        // Null if no copy constructor...
        if (proto2 != null) {
            out.println("Sequence iterator 1 (type literal):    ", proto2, " = ", ObjectPolicy.ID.toString(proto2));
            SequenceIterator<Integer> si2 = proto2.copy();
            out.println("Sequence iterator 1 copy (exact type): ", si2, " = ", ObjectPolicy.ID.toString(si2));
        }
                
        out.heading("Prototype tests succeeded!");
        
        return true;
    }
    
}
