/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 * 
 */
package dk.rode.thesis.proxy;

import static dk.rode.thesis.proxy.ProxiedPolicy.VISIBLE;

import java.lang.reflect.InvocationHandler;

import dk.rode.thesis.decorator.AppenderDecorator;
import dk.rode.thesis.meta.log.Log;
import dk.rode.thesis.meta.log.LogFactory;
import dk.rode.thesis.meta.model.EnglishAlphabetSequence;
import dk.rode.thesis.meta.model.LongSequence;
import dk.rode.thesis.meta.model.PrimeSequence;
import dk.rode.thesis.meta.model.Sequence;
import dk.rode.thesis.meta.reflect.InstantiableTypeLiteral;
import dk.rode.thesis.meta.reflect.proxy.ReferenceHandler;
import dk.rode.thesis.meta.test.Testable;
import dk.rode.thesis.meta.util.Arguments;
import dk.rode.thesis.singleton.DanishAlphabetSequence;
import dk.rode.thesis.strategy.SequencePolicy;

/**
  * <b>Proxy</b> tests.
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  */
public class Main implements Testable {

    /**
      * Executes the <b>Proxy</b> tests. 
      * 
      * @param  args The arguments to supply to {@link #test(Log, Arguments)},
      *              if any; can be null.
      * @throws Exception If the test unexpectedly fails.             
      */
    public static void main(String[] args) throws Exception {
        new Main().test(LogFactory.getLog(), new Arguments(args));
    }

    /**
      * No-arg constructor.
      */
    public Main() {
        super();
    }
    
    /**
      * 
      */
    public boolean test(Log out, Arguments arguments) throws Exception {
        out.heading("Proxy tests...");

        if (arguments.size() > 0) {
            out.println("Supplied arguments: ", arguments);
        }
       
        // True for verbose logging...
        boolean log = arguments.getArgumentValue("log.verbose", true, Boolean.class);
        // Handlers are verbose!
        LogFactory.getLog(InvocationHandler.class).setActive(log);
        LogFactory.getLog(ReferenceHandler.class).setActive(log);
        
        // True for normal class logging...
        log = arguments.getArgumentValue("log", true, Boolean.class);
        LogFactory.getLog(NonResettableSequence.class).setActive(log);
        LogFactory.getLog(ImmutableSequence.class).setActive(log);
        LogFactory.getLog(SynchronisedSequence.class).setActive(log);
        
        // Sequence proxy factory used...
        SequenceProxyFactory spf = new SequenceProxyFactory();
                
        Sequence<String> sequence = null;
        Sequence<Integer> shared = null;
        Sequence<?>[] sequences = {
                
            // Specific proxy classes
                
            // Simply proxy...
            new NonResettableSequence<String>(new EnglishAlphabetSequence()),
            // Simply proxy and its copy...
            sequence = new SynchronisedSequence<String>(new EnglishAlphabetSequence()),
            sequence.copy(),
            // Proxy to a decorator...
            new NonResettableSequence<String>(new AppenderDecorator(sequence = new EnglishAlphabetSequence(), "{", "}")),
            // The previous original sequence (set above) that will be accessed 
            // *outside* proxy for unpredictable results...
            sequence,
            // Copy that should not be affected...
            sequence = sequence.copy(),
            // Immutable access to "sequence" which should not alter it, but the
            // sequence will still be changed as it will be accessed directly
            // from the above entry...
            new ImmutableSequence<String>(sequence),
            // Proxy to proxy...
            new SynchronisedSequence<String>(new NonResettableSequence<String>(new EnglishAlphabetSequence())),
            // Non resettable singleton...
            new NonResettableSequence<String>(new SynchronisedSequence<String>(DanishAlphabetSequence.Instance)),
            // Counter...
            new LongSequence(),
            // Non resettable counter...
            new NonResettableSequence<Long>(new LongSequence(100)),
            // Fixed counter based on new sequence...
            new ImmutableSequence<Long>(new LongSequence(200)),
            
            // java.lang.reflect.Proxy proxies
            
            // Get loggable proxy for "sequence"... 
            spf.getLoggableSequence(sequence), 
            // Get synchronised proxy... 
            spf.getSynchronisedSequence(new LongSequence()),
            // Non resettable counter
            spf.getNonResettableSequence(new LongSequence(100)),
            // Immutable sequence accessing "sequence", but "sequence" is 
            // still changed elsewhere because it is registered directly 
            // as an entry in this array, i.e. accessed outside the proxy...
            sequence = spf.getImmutableSequence(sequence),
            // Try and copy it: should return the same proxy instance...
            sequence.copy(),
            // Should never be advanced, copied, or reset...
            spf.getImmutableSequence(new LongSequence(200)),
            // Proxy to proxy...
            spf.getSynchronisedSequence(sequence),
            // Virtual proxy... Notice that the construction automatically
            // handles primitives: there is no constructor in LongSequence
            // accepting a Long instance, only the primitive type long...
            spf.getVirtualSequence(new InstantiableTypeLiteral<LongSequence>(Long.class){/*magic*/}, 
                                   350 /*sequence start value*/),
            // Shared sequence...                        
            shared = spf.getSharedSequence(new PrimeSequence(1000)),
            // Copy (shared!)...
            shared = shared.copy(),
            // Copy of copy (shared!)...
            shared.copy()
        };

        // Print instances...
        out.heading("Sequence instances");
        for (int i = 0; i < sequences.length; i++) {
            // The default string representation may also be visible!
            Sequence<?> s = sequences[i];
            out.println("Sequence ", i, ": ", s).println("--> ", s.toString(VISIBLE)).println();
        }
        
        out.heading("Sequence usage");
        for (int i = 0; i < 30; i++) {
            out.println();
            out.println("Iteration: ", (i + 1));
            for (int j = 0; j < sequences.length; j++) {
                Sequence<?> s = sequences[j];
                Object value;
                if (i % 6 == 5) {
                    value = s.current();
                    s.reset();
                    out.println(j, ":\tResetting ", s, ": ", value, " -> ", s.current());
                }

                // Advance sequence...
                value = s.next();
                out.println(j, ":\t", s, " -> ", value);
                if (i % 6 == 0) {
                    out.println();
                }
            }
        }
        
        out.heading("Dynamic proxy tests");

        // Disable verbose handler logging...
        LogFactory.getLog(InvocationHandler.class).setActive(false);
        
        for (int i = 0; i < sequences.length; i++) {
            Sequence<?> s = sequences[i];
            out.println("Sequence ", i, ":\t", s.toString(SequencePolicy.ID), ": proxy: ", spf.isProxy(s), ", proxied: ", spf.isProxied(s)); 
        } 
        
        out.heading("Proxy tests succeeded!");
        return true;
    }
    
}
