/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.singleton;

import static dk.rode.thesis.meta.util.Util.add;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;

import dk.rode.thesis.meta.log.LogFactory;
import dk.rode.thesis.meta.reflect.Caller;
import dk.rode.thesis.meta.util.Util;

/**
  * A <i>mutated Simpsons family sequence</i> is a sequence representing 
  * a random mutated sub-set of the primary family members in the one and only 
  * <i>The Simpsons</i> family. <p>
  * 
  * This class can be sub-classed. <p>
  * 
  * <b>Implementation notes</b>:<br>
  * Even though the super class {@link SimpsonsFamilySequence} is a singleton 
  * class, this class is not. Creation and copying works like it does for 
  * any other non-singleton sequence. This sequence is also cloneable,
  * but will still not create additional {@code SimpsonsFamilySequence}
  * instances. <p>
  * 
  * It would in fact be possible to create {@link SimpsonsFamilySequence}
  * instances from the {@link #MutatedSimpsonsFamilySequence(int)} and
  * {@link #MutatedSimpsonsFamilySequence(MutatedSimpsonsFamilySequence)} 
  * constructors without the call to a super-constructor, i.e. using 
  * {@code new}. A {@link Caller caller} cannot determine <i>how</i> 
  * the super-class constructor is invoked from a sub-class, only 
  * from <i>where</i>. Hence, any singleton type that allows sub-classing 
  * has to trust the sub-classes to some extent. However, it is not
  * possible even for sub-classes to invoke the super constructor from
  * other places than constructors. The {@link #evilCreation()} method is
  * test method that will try and create a {@code SimpsonsFamilySequence},
  * but will fail because it is not a constructor. <p> 
  * 
  * Even though this class references static fields in 
  * {@code SimpsonsFamilySequence} class, it will not cause the
  * {@code SimpsonsFamilySequence} singleton instance to be 
  * created.
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  */
public final class MutatedSimpsonsFamilySequence extends SimpsonsFamilySequence implements Cloneable {
    
    /**
      * No-arg constructor. <p>
      * 
      * Half the members of <i>The Simpsons</i> family will be included 
      * in this mutated sequence.
      */
    public MutatedSimpsonsFamilySequence() {
        this(SimpsonsFamilySequence.SIZE / 2);
    }

    /**
      * Constructor. <p>
      * 
      * The number of members of the <i>The Simpsons</i> family included
      * in this sequence is specified by {@code members}.
      * 
      * @param  members The members; must be within 
      *                 [1, {@link SimpsonsFamilySequence#SIZE}].
      * @throws IllegalArgumentException If {@code members} is illegal.
      */
    public MutatedSimpsonsFamilySequence(int members) {
        super(getMutatedMembers(members));
        LogFactory.getLog(this).println("Created sequence: ", this, " -> ", Arrays.asList(this.elements));
    }
    
    /**
      * Copy constructor. <p>
      * 
      * @param  sequence The sequence to copy; cannot be null.
      * @throws NullPointerException If {@code sequence} is null.
      */
    public MutatedSimpsonsFamilySequence(MutatedSimpsonsFamilySequence sequence) {
        super(sequence);
    }
    
    /**
      * 
      */
    public MutatedSimpsonsFamilySequence copy() {
        return new MutatedSimpsonsFamilySequence(this);
    }

    /**
      * Tries to create a {@link SimpsonsFamilySequence} instance based
      * on an off-limit constructor. Doing so should cause  
      * a {@link SingletonError} to be thrown! If not, it
      * constitutes an internal error and the singleton invariant
      * is broken.
      *  
      * @return Should never return.
      * @throws SingletonError Always thrown. 
      */
    public SimpsonsFamilySequence evilCreation() throws SingletonError {
        return new SimpsonsFamilySequence(this.elements);    
    }
    
    /**
      * Returns a mutated sub-set of the primary family members in 
      * <i>The Simpsons</i> family of size {@code members}. 
      *  
      * @param  members The members; must be within 
      *                 [1, {@link SimpsonsFamilySequence#SIZE}].
      * @return The sub-set; never null, and never empty.
      * @throws IllegalArgumentException If {@code members} is illegal.
      */
    private final static String[] getMutatedMembers(int members) {
        if ((members < 1) || (members > SimpsonsFamilySequence.SIZE)) {
            throw new IllegalArgumentException("Illegal number of members: " + members);
        }
        List<String> list = new ArrayList<String>(SimpsonsFamilySequence.SIZE);
        add(list, SimpsonsFamilySequence.members);
        Collections.shuffle(list);
        return Util.copy(list.toArray(), new String[members], members);
    }

    /**
      * Returns a clone of this sequence. 
      * 
      * @return A clone of this sequence; never null.
      * @throws CloneNotSupportedException If the cloning fails.
      */
    public MutatedSimpsonsFamilySequence clone() throws CloneNotSupportedException {
        // Allowed because super.clone() is Object.clone which will create
        // an instance of this class, not of the super class. Invoking, for example, 
        // new SimpsonsFamilySequence(this), will throw an exception...
        return (MutatedSimpsonsFamilySequence)super.clone();
    }
    
}
