/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.singleton;

import static dk.rode.thesis.meta.util.Const.SERIAL_VERSION_UID;

/**
  * A <i>singleton exception</i> is thrown in case a given
  * type cannot be determined to be a singleton type when requested
  * from a given {@link SingletonRegistry registry} or in case
  * a singleton instance cannot be acquired from a registry. <p>
  * 
  * <b>Implementation notes</b>:<br>
  * Declared as a runtime exception because ordinary use of 
  * singletons and singleton registries are assumed to be correct 
  * by the context using them. The context, for example, should
  * never test for a null return value from a registry, because 
  * either the singleton instance exist or it does not; a null 
  * value does not make any sense, and would constitute a poor design
  * decision. Hence, in the general case, the context should not
  * have to worry about checked exceptions as they would never
  * be thrown. <p>
  * 
  * The singleton creation could fail, of cause, but that is of no
  * concern to the registry as it will simply re-throw such errors
  * wrapped in this exception type.
  *
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  */
public class SingletonException extends RuntimeException {

    /**
      * Serial version id. 
      */
    final static long serialVersionUID = SERIAL_VERSION_UID;
    
    /**
      * No-arg constructor.
      */
    public SingletonException() {
        super();
    }

    /**
      * Constructor.
      * 
      * @param type The faulting type; can be null.
      */
    public SingletonException(Class<?> type) {
        this(type == null ? null : type.getName());
    }

    /**
      * Constructor.
      * 
      * @param message The detail message; can be null.
      */
    public SingletonException(String message) {
        super(message);
    }

    /**
      * Constructor.
      * 
      * @param cause The causing throwable; can be null.
      */
    public SingletonException(Throwable cause) {
        super(cause);
    }

}
