/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.state;

import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.log.LogFactory;
import dk.rode.thesis.meta.model.AbstractSequence;
import dk.rode.thesis.meta.model.Sequence;
import dk.rode.thesis.strategy.SequencePolicy;
import dk.rode.thesis.strategy.StringablePolicy;
import dk.rode.thesis.strategy.StringablePolicy.Type;

/**
  * An <i>abstract stateable sequence</i> implements the basic
  * traits of a {@link StateableSequence stateable} sequence.
  *
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @param  <E> The type of values delivered by this stateable sequence. 
  */
@Participant("Context")
public abstract class AbstractStateableSequence<E> extends AbstractSequence<E> implements StateableSequence<E> {

    /**
      * The current <i>functional</i> sequence state encapsulating the
      * current behaviour of this sequence. <p>
      * 
      * Never null.
      */
    private FunctionalState<E> functionalState;

    /**
      * No-arg constructor. <p>
      * 
      * No initial functional state is set, and must be done in the 
      * sub-class constructor.
      */
    protected AbstractStateableSequence() {
        // do not reset, because we have not set any functional state yet...
        super(false);
    }
    
    /**
      * Constructor.
      * 
      * @param  functionalState The initial functional state of this 
      *                         sequence; cannot be null.
      * @throws NullPointerException If {@code functionalState} is null.
      */
    protected AbstractStateableSequence(FunctionalState<E> functionalState) {
        this();
        this.setFunctionalState(functionalState, State.START);
    }

    /**
      * Copy constructor.
      * 
      * @param  sequence The sequence to copy; cannot be null.
      * @throws NullPointerException If {@code sequence} is null.
      */
    protected AbstractStateableSequence(StateableSequence<E> sequence) {
        super(sequence);
        // Cannot use set method because we are not done initialising
        // in the sub-class constructor...
        this.functionalState = sequence.getFunctionalState();
        this.state = sequence.state();
    }
    
    /**
      * 
      */
    public FunctionalState<E> getFunctionalState() {
        return this.functionalState;
    }

    /**
      * 
      */
    public FunctionalState<E> setFunctionalState(FunctionalState<E> functionalState, State internalState) {
        if (functionalState == null) {
            throw new NullPointerException("Functional state cannot be null");
        }
        LogFactory.getLog(this).println("Changing functional state for ", this.toString(SequencePolicy.ID), ": [", this.functionalState, ", ", this.state, "] -> [", functionalState, ", ", internalState, "]");
        FunctionalState<E> oldState = this.functionalState;
        this.functionalState = functionalState;
        if (internalState != null) {
            this.state = internalState;
            // Take action... Since current is a null internal state,
            // we only invoke if we have a non null internal state...
            this.functionalState.action(internalState, this);
        }
        return oldState;
    }
    
    /**
      * 
      */
    public final void reset() {
        this.functionalState.action(State.RESET, this);
    }
    
    /**
      * 
      */
    public final E current() {
        return this.functionalState.action(null, this); // no action = current
    }

    /**
      * 
      */
    public final E next() {
        return this.functionalState.action(State.NORMAL, this);
    }

    /**
      * 
      */
    public CharSequence toString(StringablePolicy<? super Sequence<E>> policy) {
        policy = this.getStringablePolicy(policy);
        CharSequence cs = policy.toString(this);
        if (policy.type() == Type.CLOSED) {
            return cs;
        }
        StringBuilder sb = new StringBuilder().append(cs);
        if (policy.type() == Type.OPEN) {
            sb.append(']');
        }
        return sb.append(".[").append(this.functionalState).append(']');
    }
    
}
