/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 * 
 */
package dk.rode.thesis.strategy;

import dk.rode.thesis.bridge.SequenceAbstraction;
import dk.rode.thesis.bridge.SequenceValueRange;
import dk.rode.thesis.decorator.AppenderDecorator;
import dk.rode.thesis.decorator.UppercaseDecorator;
import dk.rode.thesis.meta.log.Log;
import dk.rode.thesis.meta.log.LogFactory;
import dk.rode.thesis.meta.model.EnglishAlphabetSequence;
import dk.rode.thesis.meta.model.LongSequence;
import dk.rode.thesis.meta.model.Sequence;
import dk.rode.thesis.meta.test.Testable;
import dk.rode.thesis.meta.util.Arguments;
import dk.rode.thesis.singleton.DanishAlphabetSequence;
import dk.rode.thesis.singleton.MutatedSimpsonsFamilySequence;
import dk.rode.thesis.state.ReversiblePrimeSequence;

/**
  * <b>Strategy</b> tests.
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  */
public class Main implements Testable {

    /**
      * Executes the <b>Strategy</b> tests.
      * 
      * @param args The arguments to supply to {@link #test(Log, Arguments)},
      *             if any; can be null.
      */
    public static void main(String[] args) {
        new Main().test(LogFactory.getLog(), new Arguments(args));
    }

    /**
      * No-arg constructor.
      */
    public Main() {
        super();
    }
    
    /**
      * 
      */
    public boolean test(Log out, Arguments arguments) {
        out.heading("Strategy tests...");
        if (arguments.size() > 0) {
            out.println("Supplied arguments: ", arguments);
        }
       
        //
        // Create sequences
        //
        
        Sequence<?>[] sequences = {
            new EnglishAlphabetSequence(),
            new UppercaseDecorator(new MutatedSimpsonsFamilySequence(), false), 
            // Cannot explicitly invoke the constructor of the Danish alphabet
            // sequence class because it is a singleton...
            new AppenderDecorator(DanishAlphabetSequence.Instance, "<", ">"),  
            DanishAlphabetSequence.Instance, 
            new SequenceAbstraction<Integer>(new SequenceValueRange(0, 10)), 
            new LongSequence(42), 
            new ReversiblePrimeSequence(42) 
        };
        
        //
        // Show sequences
        //
        
        out.heading("Created sequences");
        for (int i = 0; i < sequences.length; i++) {
            Sequence<?> sequence = sequences[i];
            // Mimic standard toString representation...
            out.println("Sequence ", i, ": ", sequence.getClass().getName(), "@", Integer.toHexString(System.identityHashCode(sequence)));
        }

        //
        // Use Sequence policies
        //
        
        out.heading("Use Sequence policies");
        
        for (SequencePolicy policy : SequencePolicy.values()) {
            out.println("Sequence policy: ", policy, " -> ", policy.type());
            for (int i = 0; i < sequences.length; i++) {
                Sequence<?> sequence = sequences[i];
                out.println("Sequence ", i, ": ", sequence.toString(policy));
            }
            out.println();
        }

        //
        // Use Object policies
        //
        
        out.heading("Use Object policies");

        for (ObjectPolicy policy : ObjectPolicy.values()) {
            out.println("Sequence policy: ", policy, " -> ", policy.type());
            for (int i = 0; i < sequences.length; i++) {
                Sequence<?> sequence = sequences[i];
                out.println("Sequence ", i, ": ", sequence.toString(policy));
            }
            out.println();
        }
        
        out.heading("Strategy tests succeeded!");
        
        return true;
    }
    
}
