/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.strategy;

import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.reflect.Reflection;

/**
  * Standard policies for formatting any type of {@code Object} 
  * into textual representations in form of char sequences (appendable) 
  * or as strings (immutable). <p>
  * 
  * <b>Implementation notes</b>:<br>
  * Though defined as an enumeration, different applicable policies 
  * may still be applied to format objects, i.e. any policy having 
  * the type {@code Stringable&ltObject>}. <p>
  * 
  * NetBeans 5.5.1 has problems with annotated enumeration constants
  * in its editor. The constants are not listed when using code-completion, 
  * merely a single {@code @} regardless of the number of constants. 
  * Even more so, the standard Java compiler version 1.6.0_01 (in NetBeans 
  * 5.5.1) does not allow <i>each constant</i> to implement interface 
  * functionality <b>without</b> the interface methods being declared 
  * (abstract) in the <i>enumeration class</i> it self, which Eclipse 
  * 3.3 does. Hence, the {@link #toString(Object)} method is declared 
  * abstract here and implemented by each constant for compatibility 
  * with the standard compiler.
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  */
public enum ObjectPolicy implements StringablePolicy<Object> {
    
    /**
      * Generates a char sequence (StringBuilder) representation of 
      * an object having the following format:
      * <pre>
      *   Class
      * </pre>
      * 
      * Where {@code Class} is the simple name of the class of 
      * the object. 
      * 
      * @see #NAME_OPEN
      */
    @Participant("ConcreteStrategy")
    NAME {
        public StringBuilder toString(Object object) {
            return new StringBuilder().append(Reflection.getClass(object.getClass()).getSimpleName());
        }
    },

    /**
      * Generates a char sequence (StringBuilder) representation of 
      * an object having the following format:
      * <pre>
      *   Class[
      * </pre>
      * 
      * Where {@code Class} is the simple name of the class of 
      * the object. <p>
      * 
      * The difference between this policy and {@link #NAME}
      * is that the representation formatted by this policy includes 
      * an opening bracket ([), i.e. this policy is considered
      * {@link StringablePolicy.Type#OPEN}. 
      */
    @Participant("ConcreteStrategy")
    NAME_OPEN {
        public StringBuilder toString(Object object) {
            return ((StringBuilder)NAME.toString(object)).append('[');
        }
        public Type type() {
            return Type.OPEN;
        }
    },
    
    /**
      * Generates a char sequence (StringBuilder) representation of 
      * an object having the following format:
      * <pre>
      *   hash-code: ..[; identity-hash-code: ..]
      * </pre>
      * 
      * Where {@code hash-code} is the hash-code of the object. <p>
      * 
      * If the <b>identity</b> hash-code differs from
      * the object hash code, {@code identity-hash-code} is the
      * identity hash code. <p>
      * 
      * Note, that identity hash codes need still not be unique
      * per unique object! <p>
      * 
      * This policy is considered {@link StringablePolicy.Type#OPEN}.
      */
    @Participant("ConcreteStrategy")
    HASHCODE {
        public StringBuilder toString(Object object) {
            int hashCode = object.hashCode(), 
                identityHashCode = System.identityHashCode(object);
            StringBuilder sb = new StringBuilder();
            sb.append("hash-code: ").append(hashCode);
            if (hashCode != identityHashCode) {
                sb.append("; identity-hash-code: ").append(identityHashCode);
            }
            return sb;
        }
        public Type type() {
            return Type.OPEN;
        }
    },
    
    /**
      * Generates a char sequence representation of an object having
      * the following format.
      * <pre>
      *   Class[..]
      * </pre>
      * 
      * Where {@code Class} is the simple class name of the class
      * of the object, and {@code ..} is the <b>identity</b> 
      * hash-code of the object.
      * 
      * @see #ID_OPEN
      */
    @Participant("ConcreteStrategy")
    ID {
        public CharSequence toString(Object object) {
            return ((StringBuilder)ID_OPEN.toString(object)).append(']');
        }
    },
    
    /**
      * Generates a char sequence (StingBuilder) representation of an 
      * object having the following format.
      * <pre>
      *   Class[..
      * </pre>
      * 
      * Where {@code Class} is the simple class name of the class
      * of the object, and {@code ..} is the <b>identity</b> 
      * hash-code of the object. <p>
      * 
      * The difference between this policy and {@link #ID}
      * is that the representation formatted by this policy does 
      * not end with a closing bracket (]), i.e. this policy is
      * {@link StringablePolicy.Type#OPEN}.
      */
    @Participant("ConcreteStrategy")
    ID_OPEN {
        public StringBuilder toString(Object object) {
            return ((StringBuilder)NAME.toString(object))
                   .append('[')
                   .append(System.identityHashCode(object));
        }
        public Type type() {
            return Type.OPEN;
        }
    },
    
    /**
      * Generates a char sequence representation of an object having
      * the following format:
      * <pre>
      *   Class[hash-code: ..[; identity-hash-code: ..]]
      * </pre>
      * 
      * Where {@code Class} is the simple class name of the class
      * of the object, and {@code hash-code} is the hash-code of
      * the object. <p>
      * 
      * If the <b>identity</b> hash-code differs from
      * the object hash code, {@code identity-hash-code} is the
      * identity hash code.
      * 
      * @see #IDENTITY_OPEN
      */
    @Participant("ConcreteStrategy")
    IDENTITY {
        public CharSequence toString(Object object) {
            return ((StringBuilder)IDENTITY_OPEN.toString(object)).append(']');
        }
    },
    
    /**
      * Generates a char sequence (StringBuilder) representation of 
      * an object having the following format:
      * <pre>
      *   Class[hash-code: ..[; identity-hash-code: ..]
      * </pre>
      * 
      * Where {@code Class} is the simple class name of the class
      * of the object, and {@code hash-code} is the hash-code of
      * the object. <p>
      * 
      * If the <b>identity</b> hash-code differs from
      * the object hash code, {@code identity-hash-code} is the
      * identity hash code. <p>
      * 
      * The difference between this policy and {@link #IDENTITY}
      * is that the representation formatted by this policy does 
      * not end with a closing bracket (]), i.e. this policy is
      * {@link StringablePolicy.Type#OPEN}.
      */
    @Participant("ConcreteStrategy")
    IDENTITY_OPEN {
        public StringBuilder toString(Object object) {
            return ((StringBuilder)NAME.toString(object))
                   .append('[')
                   .append(HASHCODE.toString(object));
        }
        public Type type() {
            return Type.OPEN;
        }
    };

    /**
      * {@inheritDoc} 
      */
    public abstract CharSequence toString(Object object);
    
    /**
      * Handy shortcut that returns a textual representation of the 
      * object supplied as {@code object} as a string. <p>
      * 
      * This is typically useful for {@link StringablePolicy.Type#CLOSED} 
      * stringable policies and/or when such a policy is applied in the standard
      * {@code Object.toString()} method.
      *  
      * @param  object The object; cannot be null.
      * @return The string representation; never null.
      * @throws NullPointerException If {@code object} is null.
      * @see    #toString(Object)
      */
    public String asString(Object object) {
        return this.toString(object).toString();
    }
    
    /**
      * @return {@link StringablePolicy.Type#APPENDABLE} is returned unless 
      *         explicitly documented otherwise for a given constant. 
      */
    public Type type() {
        return Type.APPENDABLE;
    }
        
}
