/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.visitor;

import java.util.List;

import dk.rode.thesis.composite.CompositeSequence;
import dk.rode.thesis.composite.CompositeStrategy;
import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.log.LogFactory;
import dk.rode.thesis.meta.model.Sequence;
import dk.rode.thesis.strategy.ObjectPolicy;
import dk.rode.thesis.strategy.SequencePolicy;

/**
  * A <i>simple scanner</i> can traverse both 
  * {@link TypeVisitableSequence type} and {@link ValueVisitableSequence value}
  * visitable sequences based on a traversal {@link CompositeStrategy strategy}
  * that can be supplied at {@link #SimpleScanner(CompositeStrategy) construction}
  * time, or changed runtime using the {@link #setCompositeStrategy(CompositeStrategy)}
  * method.
  *
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  */
@Participant("ObjectStructure")
public class SimpleScanner implements SequenceTypeScanner, SequenceValueScanner {

    /**
      * The composite traversal strategy to use. <p>
      * 
      * Null means default strategy per composite.
      */
    private CompositeStrategy strategy;
    
    /**
      * No-arg constructor. <p>
      * 
      * The default strategy is used per composite.
      */
    public SimpleScanner() {
        this(null);
    }

    /**
      * Constructor.
      * 
      * @param strategy The composite strategy to use for the traversal;
      *                 can be null, i.e. default.
      */
    public SimpleScanner(CompositeStrategy strategy) {
        this.setCompositeStrategy(strategy);
    }
    
    /**
      * Return the currently used {@link CompositeStrategy strategy}
      * by this scanner to traverse {@link Sequence sequence}
      * structures.
      *  
      * @return The used strategy; null means default per composite.
      * @see    #setCompositeStrategy(CompositeStrategy)
      * @see   CompositeSequence#getCompositeStrategy()
      */
    public CompositeStrategy getCompositeStrategy() {
        return this.strategy;
    }
    
    /**
      * Sets the {@link CompositeStrategy strategy} to use by this 
      * scanner to traverse {@link Sequence sequence} structures to
      * {@code strategy}
      *  
      * @param strategy The strategy to use; can be null, which means 
      *                 default strategy per composite.
      * @see   #getCompositeStrategy()
      * @see   CompositeSequence#getCompositeStrategy()
      */
    public void setCompositeStrategy(CompositeStrategy strategy) {
        this.strategy = strategy;
    }
    
    /**
      * @throws NullPointerException {@inheritDoc}
      */
    public <P> void scan(TypeVisitableSequence<?> sequence, SequenceTypeVisitor<P> visitor, P argument) {
        LogFactory.getLog(this).println("Type scan starting from ", sequence.toString(SequencePolicy.IDENTITY), ", using ", this.strategy, " strategy: ", visitor);
        
        // Composite?
        if (sequence instanceof CompositeSequence) {
            CompositeSequence<?> cs = CompositeSequence.class.cast(sequence);
            // Because "getSequences" method is declared using a class literal, not a type literal,
            // so we have to use a raw type before we can cast to TypeVisitableSequence<?> below...
            @SuppressWarnings("unchecked") List<TypeVisitableSequence> sequences = 
                cs.getSequences(this.strategy, true, null, TypeVisitableSequence.class); // true argument = include "sequence" in list...
            for (TypeVisitableSequence<?> tvs : sequences) {
                tvs.accept(visitor, argument);
            }
            
        // Single...
        } else {
            sequence.accept(visitor, argument);
        }
    }

    /**
      * @throws NullPointerException {@inheritDoc}
      */
    public <P> void scan(ValueVisitableSequence<?> sequence, SequenceValueVisitor<P> visitor, P argument) {
        LogFactory.getLog(this).println("Value scan starting from ", sequence.toString(SequencePolicy.IDENTITY), ", using ", this.strategy, " strategy: ", visitor);
        
        // Composite?
        if (sequence instanceof CompositeSequence) {
            CompositeSequence<?> cs = CompositeSequence.class.cast(sequence);
            // Because "getSequences" method is declared using a class literal, not a type literal,
            // so we have to use a raw type before we can cast to ValueVisitableSequence<?> below...
            @SuppressWarnings("unchecked") List<ValueVisitableSequence> sequences = 
                cs.getSequences(this.strategy, true, null, ValueVisitableSequence.class); // true argument = include "sequence" in list...
            for (ValueVisitableSequence<?> vvs : sequences) {
                vvs.accept(visitor, argument);
            }
            
        // Single...
        } else {
            sequence.accept(visitor, argument);
        }
    }

    /**
      * Returns the string representation of this simple scanner.
      * 
      * @return The string representation; never null.
      */
    public String toString() {
        StringBuilder sb = (StringBuilder)ObjectPolicy.NAME_OPEN.toString(this);
        return sb.append("strategy: ").append(this.strategy == null ? "default" : this.strategy.name()).append(']').toString();
    }
}
